# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import sys
import time
from enum import Enum
from copy import deepcopy

import numpy as np
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
from pyFAI.blob_detection import BlobDetection
from pyFAI.calibrant import Calibrant
from pyFAI.detectors import Detector, ALL_DETECTORS, NexusDetector
from pyFAI.geometryRefinement import GeometryRefinement
from pyFAI.massif import Massif
from skimage.measure import find_contours

from .. import calibrants_path
from .ImgModel import ImgModel
from .util import Signal
from .util.HelperModule import (
    get_base_name,
    rotate_matrix_p90,
    rotate_matrix_m90,
    get_partial_index,
)
from .util.calc import supersample_image, trim_trailing_zeros

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class CalibrationModel(object):

    def __init__(self, img_model=None):
        """
        :param img_model:
        :type img_model: ImgModel
        """
        super(CalibrationModel, self).__init__()
        self.img_model = img_model
        self.points = []
        self.points_index = []

        self.detector = Detector(pixel1=79e-6, pixel2=79e-6)
        # self.detector.shape = (2048, 2048)
        self.detector_mode = DetectorModes.CUSTOM
        self._original_detector = (
            None  # used for saving original state before rotating or flipping
        )
        self.pattern_geometry = GeometryRefinement(
            detector=self.detector, wavelength=0.3344e-10, poni1=0, poni2=0
        )  # default params are necessary, otherwise fails...
        self.pattern_geometry_img_shape = None
        self.cake_geometry = None
        self.cake_geometry_img_shape = None
        self.calibrant = Calibrant()

        self.orig_pixel1 = (
            self.detector.pixel1
        )  # needs to be extra stored for applying supersampling
        self.orig_pixel2 = self.detector.pixel2

        self.start_values = {
            "dist": 200e-3,
            "wavelength": 0.3344e-10,
            "polarization_factor": 0.99,
        }
        self.fit_wavelength = False
        self.fixed_values = (
            {}
        )  # dictionary for fixed parameters during calibration (keys can be e.g. rot1, poni1 etc.
        # and values are the values to what the respective parameter will be set
        self.is_calibrated = False
        self.use_mask = False
        self.filename = ""
        self.calibration_name = ""
        self.polarization_factor = 0.99
        self.supersampling_factor = 1
        self.correct_solid_angle = True
        self._calibrants_working_dir = calibrants_path

        self.distortion_spline_filename = None

        self.tth = np.linspace(0, 25)
        self.int = np.sin(self.tth)
        self.num_points = len(self.int)

        self.cake_img = np.zeros((2048, 2048))
        self.cake_tth = None
        self.cake_azi = None

        self.peak_search_algorithm = None

        self.img_model.img_changed.connect(self._check_detector_and_image_shape)

        self.detector_reset = Signal()

    def find_peaks_automatic(self, x, y, peak_ind):
        """
        Searches peaks by using the Massif algorithm
        :param float x:
            x-coordinate in pixel - should be from original image (not supersampled x-coordinate)
        :param float y:
            y-coordinate in pixel - should be from original image (not supersampled y-coordinate)
        :param peak_ind:
            peak/ring index to which the found points will be added
        :return:
            array of points found
        """
        massif = Massif(self.img_model.img_data, median_prefilter=False)
        cur_peak_points = massif.find_peaks(
            (int(np.round(x)), int(np.round(y))), stdout=DummyStdOut()
        )
        if len(cur_peak_points):
            self.points.append(np.array(cur_peak_points))
            self.points_index.append(peak_ind)
        return np.array(cur_peak_points)

    def find_peak(self, x, y, search_size, peak_ind):
        """
        Searches a peak around the x,y position. It just searches for the maximum value in a specific search size.
        :param int x:
            x-coordinate in pixel - should be from original image (not supersampled x-coordinate)
        :param int y:
            y-coordinate in pixel - should be form original image (not supersampled y-coordinate)
        :param search_size:
            the length of the search rectangle in pixels in all direction in which the algorithm searches for
            the maximum peak
        :param peak_ind:
            peak/ring index to which the found points will be added
        :return:
            point found (as array)
        """
        left_ind = int(np.round(x - search_size * 0.5))
        if left_ind < 0:
            left_ind = 0
        top_ind = int(np.round(y - search_size * 0.5))
        if top_ind < 0:
            top_ind = 0
        search_array = self.img_model.img_data[
            left_ind : (left_ind + search_size), top_ind : (top_ind + search_size)
        ]
        x_ind, y_ind = np.where(search_array == search_array.max())
        x_ind = x_ind[0] + left_ind
        y_ind = y_ind[0] + top_ind
        self.points.append(np.array([x_ind, y_ind]))
        self.points_index.append(peak_ind)
        return np.array([np.array((x_ind, y_ind))])

    def clear_peaks(self):
        self.points = []
        self.points_index = []

    def remove_last_peak(self):
        if self.points:
            num_points = int(
                self.points[-1].size / 2
            )  # each peak is x, y so length is twice as number of peaks
            self.points.pop(-1)
            self.points_index.pop(-1)
            return num_points

    def create_cake_geometry(self):
        self.cake_geometry = AzimuthalIntegrator(
            splineFile=self.distortion_spline_filename
        )

        pyFAI_parameter = self.pattern_geometry.getPyFAI()
        pyFAI_parameter["wavelength"] = self.pattern_geometry.wavelength

        self.cake_geometry.setPyFAI(
            dist=pyFAI_parameter["dist"],
            poni1=pyFAI_parameter["poni1"],
            poni2=pyFAI_parameter["poni2"],
            rot1=pyFAI_parameter["rot1"],
            rot2=pyFAI_parameter["rot2"],
            rot3=pyFAI_parameter["rot3"],
        )

        self.cake_geometry.detector = self.detector
        self.cake_geometry.wavelength = pyFAI_parameter["wavelength"]

    def setup_peak_search_algorithm(self, algorithm, mask=None):
        """
        Initializes the peak search algorithm on the current image
        :param algorithm:
            peak search algorithm used. Possible algorithms are 'Massif' and 'Blob'
        :param mask:
            if a mask is used during the process this is provided here as a 2d array for the image.
        """

        if algorithm == "Massif":
            self.peak_search_algorithm = Massif(
                self.img_model.img_data, median_prefilter=False
            )
        elif algorithm == "Blob":
            if mask is not None:
                self.peak_search_algorithm = BlobDetection(
                    self.img_model.img_data * mask
                )
            else:
                self.peak_search_algorithm = BlobDetection(self.img_model.img_data)
            self.peak_search_algorithm.process()
        else:
            return

    def search_peaks_on_ring(
        self, ring_index, delta_tth=0.1, min_mean_factor=1, upper_limit=55000, mask=None
    ):
        """
        This function is searching for peaks on an expected ring. It needs an initial calibration
        before. Then it will search for the ring within some delta_tth and other parameters to get
        peaks from the calibrant.

        :param ring_index: the index of the ring for the search
        :param delta_tth: search space around the expected position in two theta
        :param min_mean_factor: a factor determining the minimum peak intensity to be picked up. it is based
                                on the mean value of the search area defined by delta_tth. Pick a large value
                                for larger minimum value and lower for lower minimum value. Therefore, a smaller
                                number is more prone to picking up noise. typical values like between 1 and 3.
        :param upper_limit: maximum intensity for the peaks to be picked
        :param mask: in case the image has to be masked from certain areas, it need to be given here. Default is None.
                     The mask should be given as an 2d array with the same dimensions as the image, where 1 denotes a
                     masked pixel and all others should be 0.
        """
        self.reset_supersampling()
        if not self.is_calibrated:
            return

        # transform delta from degree into radians
        delta_tth = delta_tth / 180.0 * np.pi

        # get appropriate two theta value for the ring number
        tth_calibrant_list = self.calibrant.get_2th()
        if ring_index >= len(tth_calibrant_list):
            raise NotEnoughSpacingsInCalibrant()
        tth_calibrant = float(tth_calibrant_list[ring_index])

        # get the calculated two theta values for the whole image
        tth_array = self.pattern_geometry.twoThetaArray(self.img_model.img_data.shape)

        # create mask based on two_theta position
        ring_mask = abs(tth_array - tth_calibrant) <= delta_tth

        if mask is not None:
            mask = np.logical_and(ring_mask, np.logical_not(mask))
        else:
            mask = ring_mask

        # calculate the mean and standard deviation of this area
        sub_data = np.array(
            self.img_model.img_data.ravel()[np.where(mask.ravel())], dtype=np.float64
        )
        sub_data[np.where(sub_data > upper_limit)] = np.NaN
        mean = np.nanmean(sub_data)
        std = np.nanstd(sub_data)

        # set the threshold into the mask (don't detect very low intensity peaks)
        threshold = min_mean_factor * mean + std
        mask2 = np.logical_and(self.img_model.img_data > threshold, mask)
        mask2[np.where(self.img_model.img_data > upper_limit)] = False
        size2 = mask2.sum(dtype=int)

        keep = int(np.ceil(np.sqrt(size2)))
        try:
            old_stdout = sys.stdout
            sys.stdout = DummyStdOut
            res = self.peak_search_algorithm.peaks_from_area(
                mask2, Imin=mean - std, keep=keep
            )
            sys.stdout = old_stdout
        except IndexError:
            res = []

        # Store the result
        if len(res):
            self.points.append(np.array(res))
            self.points_index.append(ring_index)

        self.set_supersampling()
        self.pattern_geometry.reset()

    def set_calibrant(self, filename):
        self.calibrant = Calibrant()
        self.calibrant.load_file(filename)
        self.pattern_geometry.calibrant = self.calibrant

    def set_start_values(self, start_values):
        self.start_values = start_values
        self.polarization_factor = start_values["polarization_factor"]

    def set_pixel_size(self, pixel_size):
        """
        :param pixel_size: tuple with pixel_width and pixel height as element
        """
        self.orig_pixel1 = pixel_size[0]
        self.orig_pixel2 = pixel_size[1]

        self.detector.pixel1 = self.orig_pixel1
        self.detector.pixel2 = self.orig_pixel2
        self.set_supersampling()

    def update_detector_shape(self):
        self.detector.shape = self.img_model.img_data.shape
        self.detector.max_shape = self.img_model.img_data.shape

    def set_fixed_values(self, fixed_values):
        """
        Sets the fixed and not fitted values for the geometry refinement
        :param fixed_values: a dictionary with the fixed parameters as key and their corresponding fixed value, possible
                             keys: 'dist', 'rot1', 'rot2', 'rot3', 'poni1', 'poni2'

        """
        self.fixed_values = fixed_values

    def calibrate(self):
        if len(self.points) == 0:
            raise NoPointsError("No starting points for calibration found.")

        self.reset_supersampling()
        self.pattern_geometry = GeometryRefinement(
            self.create_point_array(self.points, self.points_index),
            dist=self.start_values["dist"],
            wavelength=self.start_values["wavelength"],
            detector=self.detector,
            calibrant=self.calibrant,
            splineFile=self.distortion_spline_filename,
        )

        self.refine()
        self.create_cake_geometry()
        self.is_calibrated = True

        self.calibration_name = "current"
        self.set_supersampling()
        # reset the integrator (not the geometric parameters)
        self.pattern_geometry.reset()

    def refine(self):
        if len(self.points) == 0:
            raise NoPointsError("No points for refinement found.")

        self.reset_supersampling()
        self.pattern_geometry.data = self.create_point_array(
            self.points, self.points_index
        )

        fix = ["wavelength"]
        if self.fit_wavelength:
            fix = []

        for key, value in self.fixed_values.items():
            fix.append(key)
            setattr(self.pattern_geometry, key, value)

        self.pattern_geometry.refine2(fix=fix)
        if self.fit_wavelength:
            self.pattern_geometry.refine2_wavelength(fix=fix)

        self.create_cake_geometry()
        self.set_supersampling()
        # reset the integrator (not the geometric parameters)
        self.pattern_geometry.reset()

    def _check_detector_and_image_shape(self):
        if self.detector.shape is not None:
            if self.detector.shape != self.img_model.img_data.shape:
                self.reset_detector()
                self.detector_reset.emit()
        else:
            self.reset_detector()

    def _prepare_integration_mask(self, mask):
        if mask is None:
            return self.detector.mask
        else:
            if self.detector.mask is None:
                return mask
            else:
                if mask.shape == self.detector.mask.shape:
                    return np.logical_or(self.detector.mask, mask)

    def _prepare_integration_super_sampling(self, mask):
        if self.supersampling_factor > 1:
            img_data = supersample_image(
                self.img_model.img_data, self.supersampling_factor
            )
            if mask is not None:
                mask = supersample_image(mask, self.supersampling_factor)
        else:
            img_data = self.img_model.img_data
        return img_data, mask

    def integrate_1d(
        self,
        num_points=None,
        mask=None,
        polarization_factor=None,
        filename=None,
        unit="2th_deg",
        method="csr",
        azi_range=None,
        trim_zeros=True,
    ):
        """
        :param num_points: number of points for the integration
        :param mask: mask for the integration
        :param polarization_factor: polarization factor for the integration
        :param filename: filename for saving the integration
        :param unit: unit for the integration, possible values are '2th_deg', 'q_A^-1', 'r_mm', 'r_m', 'd_A'
        :param method: method for the integration, possible values are 'csr', 'splitbbox', 'lut', 'nosplit_csr',
                          'full_csr', 'numpy', 'cython', 'BBox', 'splitPixel', 'lut_ocl', 'csr_ocl', 'csr_ocl_memsave',
                            'csr_ocl_lut', 'csr_ocl_lut_memsave', 'csr_numpy', 'csr_numpy_memsave'
        :param azi_range: azimuthal range for the integration
        :param trim_zeros: if True, the trailing zeros in the integration will be trimmed
        :return: tth, intensity
        """
        if (
            np.sum(mask)
            == self.img_model.img_data.shape[0] * self.img_model.img_data.shape[1]
        ):
            # do not perform integration if the image is completely masked...
            return self.tth, self.int

        if self.pattern_geometry_img_shape != self.img_model.img_data.shape:
            # if cake geometry was used on differently shaped image before the azimuthal integrator needs to be reset
            self.pattern_geometry.reset()
            self.pattern_geometry_img_shape = self.img_model.img_data.shape

        if polarization_factor is None:
            polarization_factor = self.polarization_factor

        self._check_detector_and_image_shape()
        mask = self._prepare_integration_mask(mask)
        img_data, mask = self._prepare_integration_super_sampling(mask)

        if num_points is None:
            num_points = self.calculate_number_of_pattern_points(img_data.shape, 2)

        self.num_points = num_points

        t1 = time.time()

        if unit == "d_A":
            try:
                self.tth, self.int = self.pattern_geometry.integrate1d(
                    img_data,
                    num_points,
                    method=method,
                    unit="2th_deg",
                    azimuth_range=azi_range,
                    mask=mask,
                    polarization_factor=polarization_factor,
                    correctSolidAngle=self.correct_solid_angle,
                    filename=filename,
                )
            except NameError:
                self.tth, self.int = self.pattern_geometry.integrate1d(
                    img_data,
                    num_points,
                    method="csr",
                    unit="2th_deg",
                    azimuth_range=azi_range,
                    mask=mask,
                    polarization_factor=polarization_factor,
                    correctSolidAngle=self.correct_solid_angle,
                    filename=filename,
                )
            self.tth = (
                self.pattern_geometry.wavelength
                / (2 * np.sin(self.tth / 360 * np.pi))
                * 1e10
            )
            self.int = self.int
        else:
            try:
                self.tth, self.int = self.pattern_geometry.integrate1d(
                    img_data,
                    num_points,
                    method=method,
                    unit=unit,
                    azimuth_range=azi_range,
                    mask=mask,
                    polarization_factor=polarization_factor,
                    correctSolidAngle=self.correct_solid_angle,
                    filename=filename,
                )
            except NameError:
                self.tth, self.int = self.pattern_geometry.integrate1d(
                    img_data,
                    num_points,
                    method="csr",
                    unit=unit,
                    azimuth_range=azi_range,
                    mask=mask,
                    polarization_factor=polarization_factor,
                    correctSolidAngle=self.correct_solid_angle,
                    filename=filename,
                )
        logger.info(
            "1d integration of {0}: {1}s.".format(
                os.path.basename(self.img_model.filename), time.time() - t1
            )
        )

        if (
            np.sum(self.int) != 0 and trim_zeros
        ):  # only trim zeros if not everything is 0 (e.g. bkg-subtraction of the same image)
            self.tth, self.int = trim_trailing_zeros(self.tth, self.int)

        return self.tth, self.int

    def integrate_2d(
        self,
        mask=None,
        polarization_factor=None,
        unit="2th_deg",
        method="csr",
        rad_points=None,
        azimuth_points=360,
        azimuth_range=None,
    ):
        if polarization_factor is None:
            polarization_factor = self.polarization_factor

        if self.cake_geometry_img_shape != self.img_model.img_data.shape:
            # if cake geometry was used on differently shaped image before the azimuthal integrator needs to be reset
            self.cake_geometry.reset()
            self.cake_geometry_img_shape = self.img_model.img_data.shape

        self._check_detector_and_image_shape()
        mask = self._prepare_integration_mask(mask)
        img_data, mask = self._prepare_integration_super_sampling(mask)

        if rad_points is None:
            rad_points = self.calculate_number_of_pattern_points(img_data.shape, 2)
        self.num_points = rad_points

        t1 = time.time()

        res = self.cake_geometry.integrate2d(
            img_data,
            rad_points,
            azimuth_points,
            azimuth_range=azimuth_range,
            method=method,
            mask=mask,
            unit=unit,
            polarization_factor=polarization_factor,
            correctSolidAngle=self.correct_solid_angle,
        )
        logger.info(
            "2d integration of {0}: {1}s.".format(
                os.path.basename(self.img_model.filename), time.time() - t1
            )
        )
        self.cake_img = res[0]
        self.cake_tth = res[1]
        self.cake_azi = res[2]
        return self.cake_img

    def cake_integral(self, tth, bins=1):
        """
        calculates a histogram of the cake in tth direction, thus the result will be pixel vs intensity
        :param tth: tth value in A^-1
        :param bins: number of bins for summing
        :return: cake_azimuth_pixel, intensity
        """
        tth_partial_index = get_partial_index(self.cake_tth, tth)
        if tth_partial_index is None:
            return [], []

        tth_center = tth_partial_index + 0.5
        left = tth_center - 0.5 * bins
        right = tth_center + 0.5 * bins

        y1 = abs(np.ceil(left) - left) * self.cake_img[:, int(np.floor(left))]
        y2 = np.sum(self.cake_img[:, int(np.ceil(left)) : int(np.floor(right))], axis=1)
        y3 = (right - np.floor(right)) * self.cake_img[:, int(np.floor(right))]

        x = np.array(range(len(self.cake_azi))) + 0.5
        y = (y1 + y2 + y3) / bins
        return x, y

    def create_point_array(self, points, points_ind):
        res = []
        for i, point_list in enumerate(points):
            if point_list.shape == (2,):
                res.append([point_list[0], point_list[1], points_ind[i]])
            else:
                for point in point_list:
                    res.append([point[0], point[1], points_ind[i]])
        return np.array(res)

    def get_point_array(self):
        return self.create_point_array(self.points, self.points_index)

    def get_calibration_parameter(self):
        pyFAI_parameter = self.pattern_geometry.getPyFAI()
        pyFAI_parameter["polarization_factor"] = self.polarization_factor
        try:
            fit2d_parameter = self.pattern_geometry.getFit2D()
            fit2d_parameter["polarization_factor"] = self.polarization_factor
        except TypeError:
            fit2d_parameter = None

        pyFAI_parameter["wavelength"] = self.pattern_geometry.wavelength
        if fit2d_parameter:
            fit2d_parameter["wavelength"] = self.pattern_geometry.wavelength

        return pyFAI_parameter, fit2d_parameter

    def calculate_number_of_pattern_points(self, img_shape, max_dist_factor=1.5):
        # calculates the number of points for an integrated pattern, based on the distance of the beam center to the the
        # image corners. Maximum value is determined by the shape of the image.
        fit2d_parameter = self.pattern_geometry.getFit2D()
        center_x = fit2d_parameter["centerX"]
        center_y = fit2d_parameter["centerY"]
        width, height = img_shape

        if width > center_x > 0:
            side1 = np.max([abs(width - center_x), center_x])
        else:
            side1 = width

        if center_y < height and center_y > 0:
            side2 = np.max([abs(height - center_y), center_y])
        else:
            side2 = height
        max_dist = np.sqrt(side1**2 + side2**2)
        return int(max_dist * max_dist_factor)

    def load(self, filename):
        """
        Loads a calibration file andsets all the calibration parameter.
        :param filename: filename for a *.poni calibration file
        """
        self.pattern_geometry = GeometryRefinement(
            wavelength=0.3344e-10, detector=self.detector, poni1=0, poni2=0
        )  # default params are necessary, otherwise fails...
        self.pattern_geometry.load(filename)
        self.orig_pixel1 = self.pattern_geometry.pixel1
        self.orig_pixel2 = self.pattern_geometry.pixel2

        if (
            self.pattern_geometry.pixel1 == self.detector.pixel1
            and self.pattern_geometry.pixel2 == self.detector.pixel2
        ):
            self.pattern_geometry.detector = (
                self.detector
            )  # necessary since loading a poni file will reset the detector
        else:
            self.reset_detector()

        self.calibration_name = get_base_name(filename)
        self.filename = filename
        self.is_calibrated = True
        self.create_cake_geometry()
        self.set_supersampling()

    def save(self, filename):
        """
        Saves the current calibration parameters into a a text file. Default extension is
        *.poni
        """
        self.cake_geometry.save(filename)
        self.calibration_name = get_base_name(filename)
        self.filename = filename

    def load_detector(self, name):
        self.detector_mode = DetectorModes.PREDEFINED
        names, classes = get_available_detectors()
        detector_ind = names.index(name)

        self._load_detector(classes[detector_ind]())

    def load_detector_from_file(self, filename):
        self.detector_mode = DetectorModes.NEXUS
        self._load_detector(NexusDetector(filename))

    def _load_detector(self, detector):
        """Loads a pyFAI detector
        :param detector: an instance of pyFAI Detector
        """
        self.detector = detector
        self.detector.calc_mask()
        self.orig_pixel1 = self.detector.pixel1
        self.orig_pixel2 = self.detector.pixel2

        self.pattern_geometry.detector = self.detector

        if self.cake_geometry:
            self.cake_geometry.detector = self.detector

        self.set_supersampling()
        self._original_detector = None

    def reset_detector(self):
        self.detector_mode = DetectorModes.CUSTOM
        self.detector = Detector(
            pixel1=self.detector.pixel1, pixel2=self.detector.pixel2
        )
        self.update_detector_shape()
        self.pattern_geometry.detector = self.detector
        if self.cake_geometry:
            self.cake_geometry.detector = self.detector
        self.set_supersampling()

    def create_file_header(self):
        try:
            # pyFAI version 0.12.0
            return self.pattern_geometry.makeHeaders(
                polarization_factor=self.polarization_factor
            )
        except AttributeError:
            # pyFAI after version 0.12.0
            from pyFAI.io import DefaultAiWriter

            return DefaultAiWriter(None, self.pattern_geometry).make_headers()

    def set_fit2d(self, fit2d_parameter):
        """
        Reads in a dictionary with fit2d parameters where the fields of the dictionary are:
        'directDist', 'centerX', 'centerY', 'tilt', 'tiltPlanRotation', 'pixelX', pixelY',
        'polarization_factor', 'wavelength'
        """
        self.pattern_geometry.setFit2D(
            directDist=fit2d_parameter["directDist"],
            centerX=fit2d_parameter["centerX"],
            centerY=fit2d_parameter["centerY"],
            tilt=fit2d_parameter["tilt"],
            tiltPlanRotation=fit2d_parameter["tiltPlanRotation"],
            pixelX=fit2d_parameter["pixelX"],
            pixelY=fit2d_parameter["pixelY"],
        )
        # the detector pixel1 and pixel2 values are updated by setPyFAI
        self.orig_pixel1 = self.detector.pixel1
        self.orig_pixel2 = self.detector.pixel2

        self.pattern_geometry.wavelength = fit2d_parameter["wavelength"]
        self.create_cake_geometry()
        self.polarization_factor = fit2d_parameter["polarization_factor"]
        self.orig_pixel1 = fit2d_parameter["pixelX"] * 1e-6
        self.orig_pixel2 = fit2d_parameter["pixelY"] * 1e-6
        self.is_calibrated = True
        self.set_supersampling()

    def set_pyFAI(self, pyFAI_parameter):
        """
        Reads in a dictionary with pyFAI parameters where the fields of dictionary are:
        'dist', 'poni1', 'poni2', 'rot1', 'rot2', 'rot3', 'pixel1', 'pixel2', 'wavelength',
        'polarization_factor'
        """
        self.pattern_geometry.setPyFAI(
            dist=pyFAI_parameter["dist"],
            poni1=pyFAI_parameter["poni1"],
            poni2=pyFAI_parameter["poni2"],
            rot1=pyFAI_parameter["rot1"],
            rot2=pyFAI_parameter["rot2"],
            rot3=pyFAI_parameter["rot3"],
            pixel1=pyFAI_parameter["pixel1"],
            pixel2=pyFAI_parameter["pixel2"],
        )
        self.detector.pixel1 = pyFAI_parameter["pixel1"]
        self.detector.pixel2 = pyFAI_parameter["pixel2"]
        self.orig_pixel1 = self.detector.pixel1
        self.orig_pixel2 = self.detector.pixel2
        self.pattern_geometry.wavelength = pyFAI_parameter["wavelength"]
        self.create_cake_geometry()
        self.polarization_factor = pyFAI_parameter["polarization_factor"]
        self.is_calibrated = True
        self.set_supersampling()

    def load_distortion(self, spline_filename):
        self.distortion_spline_filename = spline_filename
        self.pattern_geometry.set_splineFile(spline_filename)
        if self.cake_geometry:
            self.cake_geometry.set_splineFile(spline_filename)

    def reset_distortion_correction(self):
        self.distortion_spline_filename = None
        self.detector.set_splineFile(None)
        self.pattern_geometry.set_splineFile(None)
        if self.cake_geometry:
            self.cake_geometry.set_splineFile(None)

    def set_supersampling(self, factor=None):
        """
        Sets the supersampling to a specific factor. Whereby the factor determines in how many artificial pixel the
        original pixel is split. (factor^2)

        factor  n_pixel
        1       1
        2       4
        3       9
        4       16
        """
        if factor is None:
            factor = self.supersampling_factor

        self.detector.pixel1 = self.orig_pixel1 / float(factor)
        self.detector.pixel2 = self.orig_pixel2 / float(factor)
        self.pattern_geometry.pixel1 = self.orig_pixel1 / float(factor)
        self.pattern_geometry.pixel2 = self.orig_pixel2 / float(factor)

        if factor != self.supersampling_factor:
            self.pattern_geometry.reset()
            self.supersampling_factor = factor

    def reset_supersampling(self):
        self.pattern_geometry.pixel1 = self.orig_pixel1
        self.pattern_geometry.pixel2 = self.orig_pixel2
        self.detector.pixel1 = self.orig_pixel1
        self.detector.pixel2 = self.orig_pixel2

    def get_two_theta_img(self, x, y):
        """
        Gives the two_theta value for the x,y coordinates on the image. Be aware that this function will be incorrect
        for pixel indices, since it does not correct for center of the pixel.
        :param  x: x-coordinate in pixel on the image
        :type   x: ndarray
        :param  y: y-coordinate in pixel on the image
        :type   y: ndarray

        :return  : two theta in radians
        """
        if not isinstance(x, np.ndarray):
            x = np.array([x])
        if not isinstance(y, np.ndarray):
            y = np.array([y])
        x *= self.supersampling_factor
        y *= self.supersampling_factor

        return self.pattern_geometry.tth(x - 0.5, y - 0.5)[
            0
        ]  # deletes 0.5 because tth function uses pixel indices

    def get_azi_img(self, x, y):
        """
        Gives chi for position on image.
        :param  x: x-coordinate in pixel on the image
        :type   x: ndarray
        :param  y: y-coordinate in pixel on the image
        :type   y: ndarray

        :return  : azimuth in radians
        """
        # if float convert to np.array:
        if not isinstance(x, np.ndarray):
            x = np.array([x])
        if not isinstance(y, np.ndarray):
            y = np.array([y])
        x *= self.supersampling_factor
        y *= self.supersampling_factor
        return self.pattern_geometry.chi(x - 0.5, y - 0.5)[0]

    def get_two_theta_array(self):
        return self.pattern_geometry.twoThetaArray(self.img_model.img_data.shape)[
            :: self.supersampling_factor, :: self.supersampling_factor
        ]

    def get_pixel_ind(self, tth, azi):
        """
        Calculates pixel index for a specfic two theta and azimutal value.
        :param tth:
            two theta in radians
        :param azi:
            azimuth in radians
        :return:
            tuple of index 1 and 2
        """
        tth_ind = find_contours(self.pattern_geometry.ttha, tth)
        if len(tth_ind) == 0:
            return []
        tth_ind = np.vstack(tth_ind)
        azi_values = self.pattern_geometry.chi(tth_ind[:, 0], tth_ind[:, 1])
        min_index = np.argmin(np.abs(azi_values - azi))
        return tth_ind[min_index, 0], tth_ind[min_index, 1]

    @property
    def wavelength(self):
        return self.pattern_geometry.wavelength

    ##########################
    ## Detector rotation stuff
    def swap_detector_shape(self):
        self._swap_detector_shape()
        self._swap_pixel_size()
        self._swap_detector_module_size()

    def rotate_detector_m90(self):
        """
        Rotates the detector stuff by m90 degree. This includes swapping of shape, pixel size and module sizes, as well
        as dx and dy
        """
        self._save_original_detector_definition()

        self.swap_detector_shape()
        self._reset_detector_mask()
        self._transform_pixel_corners(rotate_matrix_m90)

    def rotate_detector_p90(self):
        """ """
        self._save_original_detector_definition()

        self.swap_detector_shape()
        self._reset_detector_mask()
        self._transform_pixel_corners(rotate_matrix_p90)

    def flip_detector_horizontally(self):
        self._save_original_detector_definition()
        self._transform_pixel_corners(np.fliplr)

    def flip_detector_vertically(self):
        self._save_original_detector_definition()
        self._transform_pixel_corners(np.flipud)

    def reset_transformations(self):
        """Restores the detector to it's original state"""
        if self._original_detector is None:  # no transformations done so far
            return

        self.detector = deepcopy(self._original_detector)
        self.orig_pixel1, self.orig_pixel2 = self.detector.pixel1, self.detector.pixel2
        self.pattern_geometry.detector = self.detector
        if self.cake_geometry is not None:
            self.cake_geometry.detector = self.detector
        self.set_supersampling()
        self._original_detector = None

    def load_transformations_string_list(self, transformations):
        """Transforms the detector parameters (shape, pixel size and distortion correction) based on a
        list of transformation actions.
        :param transformations: list of transformations specified as strings, values are "flipud", "fliplr",
                                "rotate_matrix_m90", "rotate_matrix_p90
        """
        for transformation in transformations:
            if transformation == "flipud":
                self.flip_detector_vertically()
            elif transformation == "fliplr":
                self.flip_detector_horizontally()
            elif transformation == "rotate_matrix_m90":
                self.rotate_detector_m90()
            elif transformation == "rotate_matrix_p90":
                self.rotate_detector_p90()

    def _save_original_detector_definition(self):
        """
        Saves the state of the detector to _original_detector if not done yet. Used for restoration upon resetting
        the transformations.
        """
        if self._original_detector is None:
            self._original_detector = deepcopy(self.detector)
            self._original_detector.pixel1 = self.orig_pixel1
            self._original_detector.pixel2 = self.orig_pixel2
            self._mask = False

    def _transform_pixel_corners(self, transform_function):
        """
        :param transform_function: function pointer which will affect the dx, dy and pixel corners of the detector
        """
        if self.detector._pixel_corners is not None:
            self.detector._pixel_corners = np.ascontiguousarray(
                transform_function(self.detector.get_pixel_corners())
            )

    def _swap_pixel_size(self):
        """swaps the pixel sizes"""
        self.orig_pixel1, self.orig_pixel2 = self.orig_pixel2, self.orig_pixel1
        self.set_supersampling()

    def _swap_detector_shape(self):
        """Swaps the detector shape and max_shape values"""
        if self.detector.shape is not None:
            self.detector.shape = (self.detector.shape[1], self.detector.shape[0])
        if self.detector.max_shape is not None:
            self.detector.max_shape = (
                self.detector.max_shape[1],
                self.detector.max_shape[0],
            )

    def _swap_detector_module_size(self):
        """swaps the module size and gap sizes for e.g. Pilatus Detectors"""
        if hasattr(self.detector, "module_size"):
            self.detector.module_size = (
                self.detector.module_size[1],
                self.detector.module_size[0],
            )
        if hasattr(self.detector, "MODULE_GAP"):
            self.detector.MODULE_GAP = (
                self.detector.MODULE_GAP[1],
                self.detector.MODULE_GAP[0],
            )

    def _reset_detector_mask(self):
        """resets and recalculates the mask. Transformations to shape and module size have to be performed before."""
        self.detector._mask = False


class DetectorModes(Enum):
    CUSTOM = 1
    NEXUS = 2
    PREDEFINED = 3


class NotEnoughSpacingsInCalibrant(Exception):
    pass


class DetectorShapeError(Exception):
    pass


class NoPointsError(Exception):
    pass


class DummyStdOut(object):
    @classmethod
    def write(cls, *args, **kwargs):
        pass


def get_available_detectors():
    detector_classes = set()
    detector_names = []

    for key, item in ALL_DETECTORS.items():
        detector_classes.add(item)

    for detector in detector_classes:
        if len(detector.aliases) > 0:
            detector_names.append(detector.aliases[0])
        else:
            detector_names.append(detector().__class__.__name__)

    sorted_indices = sorted(range(len(detector_names)), key=detector_names.__getitem__)

    detector_names_sorted = [detector_names[i] for i in sorted_indices]
    detector_classes_sorted = [list(detector_classes)[i] for i in sorted_indices]

    base_class_index = detector_names_sorted.index("Detector")
    del detector_names_sorted[base_class_index]
    del detector_classes_sorted[base_class_index]

    return detector_names_sorted, detector_classes_sorted



from ....model.util.HelperModule import get_partial_index

# imports for type hinting in PyCharm -- DO NOT DELETE
from ....model.DioptasModel import DioptasModel
from ....widgets.integration import BatchWidget
from ....widgets.plot_widgets.ImgWidget import IntegrationImgWidget


class PhaseInBatchController(object):
    """
    PhaseInBatchController handles all the interaction between the phase controls and the plotted lines in the cake view.
    """

    def __init__(self, batch_widget: BatchWidget, dioptas_model: DioptasModel):
        """
        :param batch_widget: Reference to an IntegrationWidget
        :param dioptas_model: reference to DioptasModel object
        """
        self.model = dioptas_model
        self.phase_model = self.model.phase_model
        self.batch_widget = batch_widget
        self.batch_view_widget = batch_widget.stack_plot_widget.img_view  # type: IntegrationImgWidget

        self.connect()

    def connect(self):
        self.phase_model.phase_added.connect(self.add_phase_plot)
        self.model.phase_model.phase_removed.connect(self.batch_view_widget.del_cake_phase)

        self.phase_model.phase_changed.connect(self.update_phase_lines)
        self.phase_model.phase_changed.connect(self.update_phase_color)
        self.phase_model.phase_changed.connect(self.update_phase_visible)

        self.phase_model.reflection_added.connect(self.reflection_added)
        self.phase_model.reflection_deleted.connect(self.reflection_deleted)

        self.model.enabled_phases_in_cake.connect(self.update_all_phases)

    def get_phase_position_and_intensities(self, ind, clip=True):
        """
        Obtains the positions and intensities for lines of a phase with an index ind within the batch view.

        No clipping is used for the first call to add the CakePhasePlot to the ImgWidget. Subsequent calls are used with
        clipping. Thus, only lines within the cake_tth are returned. The visibility of each line is then estimated in
        the ImgWidget based on the length of the clipped and not clipped lists.

        :param ind: the index of the phase
        :param clip: whether or not the lists should be clipped. Clipped means that lines which have positions larger
                     than the
        :return: line_positions, line_intensities
        """
        if self.model.batch_model.binning is None:
            cake_tth = self.model.calibration_model.tth
        else:
            start_x, stop_x = self._get_x_range()
            cake_tth = self.model.batch_model.binning[start_x:stop_x]
        reflections_tth = self.phase_model.get_phase_line_positions(ind, '2th_deg',
                                                                    self.model.calibration_model.wavelength * 1e10)
        reflections_intensities = [reflex[1] for reflex in self.phase_model.reflections[ind]]

        cake_line_positions = []
        cake_line_intensities = []

        for ind, tth in enumerate(reflections_tth):
            pos_ind = get_partial_index(cake_tth, tth)
            if pos_ind is not None:
                cake_line_positions.append(pos_ind + 0.5)
                cake_line_intensities.append(reflections_intensities[ind])
            elif clip is False:
                cake_line_positions.append(0)
                cake_line_intensities.append(reflections_intensities[ind])

        return cake_line_positions, cake_line_intensities

    def add_phase_plot(self):
        cake_line_positions, cake_line_intensities = self.get_phase_position_and_intensities(-1, False)

        self.batch_view_widget.add_cake_phase(cake_line_positions, cake_line_intensities,
                                              self.phase_model.phase_colors[-1])

    def update_phase_lines(self, ind):
        cake_line_positions, cake_line_intensities = self.get_phase_position_and_intensities(ind)
        self.batch_view_widget.update_phase_intensities(ind, cake_line_positions, cake_line_intensities)

    def update_all_phases(self):
        for ind in range(len(self.phase_model.phases)):
            self.update_phase_lines(ind)

    def update_phase_color(self, ind):
        self.batch_view_widget.set_cake_phase_color(ind, self.model.phase_model.phase_colors[ind])

    def update_phase_visible(self, ind):
        if self.phase_model.phase_visible[ind] and self.batch_widget.control_widget.phases_btn.isChecked():
            self.batch_view_widget.show_cake_phase(ind)
        else:
            self.batch_view_widget.hide_cake_phase(ind)

    def reflection_added(self, ind):
        self.batch_view_widget.phases[ind].add_line()

    def reflection_deleted(self, phase_ind, reflection_ind):
        self.batch_view_widget.phases[phase_ind].delete_line(reflection_ind)

    def _get_x_range(self):
        """
        Return bin-x range of the batch plot
        """
        if self.model.batch_model.data is None:
            return 0, 0
        start_x = 0
        stop_x = self.model.batch_model.data.shape[1]
        if self.batch_widget.options_widget.bkg_cut_btn.isChecked():
            bkg_roi = self.model.pattern_model.pattern.auto_background_subtraction_roi
            if bkg_roi is not None:
                binning = self.model.batch_model.binning
                scale = (binning[-1] - binning[0]) / binning.shape[0]
                start_x = (bkg_roi[0] - binning[0]) / scale
                stop_x = (bkg_roi[1] - binning[0]) / scale
        return int(start_x), int(stop_x)



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from qtpy import QtCore, QtWidgets, QtGui
import os
from .CustomWidgets import FlatButton
from .. import style_path


class CifConversionParametersDialog(QtWidgets.QDialog):
    """
    Dialog which is asking for Intensity Cutoff and minimum d-spacing when loading cif files.
    """

    def __init__(self, parent):
        super(CifConversionParametersDialog, self).__init__()

        self._parent = parent
        self._create_widgets()
        self._layout_widgets()
        self._style_widgets()

        self._connect_widgets()

    def _create_widgets(self):
        """
        Creates all necessary widgets.
        """
        self.int_cutoff_lbl = QtWidgets.QLabel("Intensity Cutoff:")
        self.min_d_spacing_lbl = QtWidgets.QLabel("Minimum d-spacing:")

        self.int_cutoff_txt = QtWidgets.QLineEdit("0.5")
        self.int_cutoff_txt.setToolTip(
            "Reflections with lower Intensity won't be considered."
        )
        self.min_d_spacing_txt = QtWidgets.QLineEdit("0.5")
        self.min_d_spacing_txt.setToolTip(
            "Reflections with smaller d_spacing won't be considered."
        )

        self.int_cutoff_unit_lbl = QtWidgets.QLabel("%")
        self.min_d_spacing_unit_lbl = QtWidgets.QLabel("A")

        self.ok_btn = FlatButton("OK")

    def _layout_widgets(self):
        """
        Layouts the widgets into a gridlayout
        """
        self._layout = QtWidgets.QGridLayout()
        self._layout.addWidget(self.int_cutoff_lbl, 0, 0)
        self._layout.addWidget(self.int_cutoff_txt, 0, 1)
        self._layout.addWidget(self.int_cutoff_unit_lbl, 0, 2)
        self._layout.addWidget(self.min_d_spacing_lbl, 1, 0)
        self._layout.addWidget(self.min_d_spacing_txt, 1, 1)
        self._layout.addWidget(self.min_d_spacing_unit_lbl, 1, 2)
        self._layout.addWidget(self.ok_btn, 2, 1, 1, 2)

        self.setLayout(self._layout)

    def _style_widgets(self):
        """
        Makes everything pretty and set Double validators for the line edits.
        """
        self.int_cutoff_lbl.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.int_cutoff_txt.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.int_cutoff_txt.setMaximumWidth(40)
        self.min_d_spacing_lbl.setAlignment(
            QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter
        )
        self.min_d_spacing_txt.setAlignment(
            QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter
        )
        self.min_d_spacing_txt.setMaximumWidth(40)

        self.int_cutoff_txt.setValidator(QtGui.QDoubleValidator())
        self.min_d_spacing_txt.setValidator(QtGui.QDoubleValidator())

        self.setWindowFlags(QtCore.Qt.FramelessWindowHint)

        file = open(os.path.join(style_path, "stylesheet.qss"))
        stylesheet = file.read()
        self.setStyleSheet(stylesheet)
        file.close()

    def _connect_widgets(self):
        """
        Connecting actions to slots.
        """
        self.ok_btn.clicked.connect(self.accept)

    @property
    def int_cutoff(self):
        """
        Returns the intensity cutoff selected in the dialog.
        """
        return float(str(self.int_cutoff_txt.text()))

    @property
    def min_d_spacing(self):
        """
        Returns the minimum d-spacing selected in the dialog.
        """
        return float(str(self.min_d_spacing_txt.text()))

    def exec_(self):
        """
        Overwriting the dialog exec_ function to center the widget in the parent window before execution.
        """
        parent_center = self._parent.window().mapToGlobal(
            self._parent.window().rect().center()
        )
        self.move(parent_center.x() - 101, parent_center.y() - 48)
        super(CifConversionParametersDialog, self).exec_()


class FileInfoWidget(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super(FileInfoWidget, self).__init__(parent)
        self.setWindowTitle("File Info")

        self.text_lbl = QtWidgets.QLabel()
        self.text_lbl.setWordWrap(True)

        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(5, 5, 5, 5)
        self._layout.addWidget(self.text_lbl)
        self._layout.setSizeConstraint(QtWidgets.QLayout.SetFixedSize)

        self.setStyleSheet(
            """
            QWidget{
                background: rgb(0,0,0);
            }
            QLabel{
                color: #00DD00;
            }"""
        )
        self.setLayout(self._layout)
        self.setWindowFlags(QtCore.Qt.Tool | QtCore.Qt.MSWindowsFixedSizeDialogHint)
        self.setAttribute(QtCore.Qt.WA_MacAlwaysShowToolWindow)

    def raise_widget(self):
        self.show()
        self.setWindowState(
            self.windowState() & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive
        )
        self.activateWindow()
        self.raise_()


class ErrorMessageBox(QtWidgets.QDialog):
    def __init__(self, *args, **kwargs):
        super(ErrorMessageBox, self).__init__(*args, **kwargs)
        self.setWindowTitle("OOOPS! An error occurred!")

        self.text_lbl = QtWidgets.QLabel()
        self.text_lbl.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        self.scroll_area = QtWidgets.QScrollArea()

        self.scroll_area.setWidget(self.text_lbl)
        self.scroll_area.setWidgetResizable(True)
        self.ok_btn = QtWidgets.QPushButton("OK")

        _layout = QtWidgets.QGridLayout()
        _layout.addWidget(self.scroll_area, 0, 0, 1, 10)
        _layout.addWidget(self.ok_btn, 1, 9)

        self.setLayout(_layout)
        self.ok_btn.clicked.connect(self.close)

    def setText(self, text_str):
        self.text_lbl.setText(text_str)


def open_file_dialog(parent_widget, caption, directory, filter=None):
    filename = QtWidgets.QFileDialog.getOpenFileName(
        parent_widget, caption=caption, directory=directory, filter=filter
    )
    if isinstance(filename, tuple):  # PyQt5 returns a tuple...
        return str(filename[0])
    return str(filename)


def open_files_dialog(parent_widget, caption, directory, filter=None):
    filenames = QtWidgets.QFileDialog.getOpenFileNames(
        parent_widget, caption=caption, directory=directory, filter=filter
    )
    if isinstance(filenames, tuple):  # PyQt5 returns a tuple...
        filenames = filenames[0]
    return filenames


def save_file_dialog(parent_widget, caption, directory, filter=None):
    filename = QtWidgets.QFileDialog.getSaveFileName(
        parent_widget, caption, directory=directory, filter=filter
    )
    if isinstance(filename, tuple):  # PyQt5 returns a tuple...
        return set_extension(str(filename[0]), str(filename[1]))
    return str(filename)


def set_extension(filename, ext_filter):
    """
    Set extension to filename if not given.
    Extension is taken from filter string.

    :param filename: Name of file
    :param ext_filter: Filter string used in getSaveFileName dialog
    """
    name, ext = os.path.splitext(filename)

    # extension already given
    if ext != "":
        return filename

    # Extension can not be extracted from filter string
    if ext_filter.count("(") != 1:
        return filename

    ext = ext_filter[ext_filter.find("(") + 2 : ext_filter.find(")")]
    return f"{filename}{ext}"


def get_progress_dialog(
    message: str,
    abort_text: str,
    num_points: int,
    parent=None,
) -> QtWidgets.QProgressDialog:
    progress_dialog = QtWidgets.QProgressDialog(
        message, abort_text, 0, int(num_points), parent=parent
    )
    progress_dialog.setWindowModality(QtCore.Qt.WindowModal)
    progress_dialog.setWindowFlags(QtCore.Qt.FramelessWindowHint)
    pos = parent.rect().center()
    progress_dialog.move(
        int(pos.x() - progress_dialog.width() / 2), int(pos.y() - progress_dialog.height() / 2)
    )
    progress_dialog.show()
    return progress_dialog



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional
import logging
from copy import copy

import numpy as np
from xypattern import Pattern

from .util.HelperModule import calculate_color, rgb_to_hex
from .util import Signal

logger = logging.getLogger(__name__)


class Overlay(Pattern):
    """
    Overlay class, inherits from Pattern. It is used to store overlays for the pattern widget.
    """

    index = 0

    def __init__(
        self, x: np.ndarray | None = None, y: np.ndarray | None = None, name: str = ""
    ):
        super(Overlay, self).__init__(x, y, name)
        self.visible = True
        self.color = rgb_to_hex(calculate_color(Overlay.index))
        Overlay.index += 1

    @classmethod
    def from_pattern(cls, pattern: Pattern):
        """
        Creates an overlay from a pattern, does not use its original scaling parameters
        """
        return Overlay(np.copy(pattern.x), np.copy(pattern.y), copy(pattern.name))


class OverlayModel(object):
    """
    Main Overlay Pattern Handling Model. (Was previously included in the PatternModel),
    """

    def __init__(self):
        super(OverlayModel, self).__init__()
        self.overlays = []

        self.overlay_added = Signal()
        self.overlay_changed = Signal(int)  # changed index
        self.overlay_removed = Signal(int)  # removed index

    def add_overlay(self, x: np.ndarray, y: np.ndarray, name: str = ""):
        """
        Adds an overlay to the list of overlays
        :param x: x-values
        :param y: y-values
        :param name: name of overlay to be used for displaying etc.
        """
        self.add_overlay_pattern(Overlay(x, y, name))
        return self.overlays[-1]

    def add_overlay_pattern(self, pattern: Pattern):
        """
        Adds a pattern as overlay to the list of overlays, does not use its original scaling parameters
        """
        overlay_pattern = Overlay.from_pattern(pattern)
        self.overlays.append(overlay_pattern)
        self.overlay_added.emit()

    def add_overlay_file(self, filename: str):
        """
        Reads a 2-column (x,y) text file and adds it as overlay to the list of overlays
        :param filename: path of the file to be loaded
        """
        pattern = Overlay.from_file(filename)
        self.add_overlay_pattern(pattern)

    def remove_overlay(self, ind: int):
        """
        Removes an overlay from the list of overlays
        :param ind: index of the overlay
        """
        if ind >= 0:
            del self.overlays[ind]
            self.overlay_removed.emit(ind)

    def get_overlay(self, ind: int) -> Optional[Overlay]:
        """
        :param ind: overlay ind
        :return: returns overlay if existent or None if it does not exist
        """
        try:
            return self.overlays[ind]
        except IndexError:
            return None

    def move_overlay_up(self, ind: int):
        """
        Moves the overlay up in the list of overlays (i.e. from positon 3 to 2)
        :param ind: index of the overlay
        """
        if ind > 0:
            self.overlays[ind], self.overlays[ind - 1] = (
                self.overlays[ind - 1],
                self.overlays[ind],
            )
            self.overlay_changed.emit(ind)
            self.overlay_changed.emit(ind - 1)

    def move_overlay_down(self, ind: int):
        """
        Moves the overlay down in the list of overlays (i.e. from position 3 to
        4)
        :param ind: index of the overlay
        """
        if ind < len(self.overlays) - 1:
            self.overlays[ind], self.overlays[ind + 1] = (
                self.overlays[ind + 1],
                self.overlays[ind],
            )
            self.overlay_changed.emit(ind)
            self.overlay_changed.emit(ind + 1)

    def set_overlay_scaling(self, ind: int, scaling: float):
        """
        Sets the scaling of the specified overlay
        :param ind: index of the overlay
        :param scaling: new scaling value
        """
        self.overlays[ind].scaling = scaling
        self.overlay_changed.emit(ind)

    def get_overlay_scaling(self, ind: int) -> float:
        """
        Returns the scaling of the specified overlay
        :param ind: index of the overlay
        :return: scaling value
        """
        return self.overlays[ind].scaling

    def set_overlay_offset(self, ind: int, offset: float):
        """
        Sets the offset of the specified overlay
        :param ind: index of the overlay
        :param offset: new offset value
        """
        self.overlays[ind].offset = offset
        self.overlay_changed.emit(ind)

    def get_overlay_offset(self, ind: int) -> float:
        """
        Return the offset of the specified overlay
        :param ind: index of the overlay
        :return: overlay value
        """
        return self.overlays[ind].offset

    def set_overlay_visible(self, ind: int, visible: bool):
        """
        Sets the visibility of the specified overlay
        :param ind: index of the overlay
        :param visible: new visibility value (True or False)
        """
        self.overlays[ind].visible = visible
        self.overlay_changed.emit(ind)

    def set_overlay_color(self, ind: int, color: str):
        """
        Sets the color of the specified overlay (as hex string, e.g. #FF0000)
        :param ind: index of the overlay
        :param color: new color value
        """
        self.overlays[ind].color = color
        self.overlay_changed.emit(ind)

    def get_overlay_color(self, ind: int) -> str:
        """
        Returns the color of the specified overlay
        :param ind: index of the overlay
        :return: color as hex string (e.g. #FF0000)
        """
        return self.overlays[ind].color

    def set_overlay_name(self, ind: int, name: str):
        """
        Sets the name of the specified overlay
        :param ind: index of the overlay
        :param name: new name value
        """
        self.overlays[ind].name = name
        self.overlay_changed.emit(ind)

    def get_overlay_name(self, ind: int) -> str:
        """
        Returns the name of the specified overlay
        :param ind: index of the overlay
        :return: name value
        """
        return self.overlays[ind].name

    def overlay_waterfall(self, separation: float):
        offset = 0
        for ind in range(len(self.overlays)):
            offset -= separation
            self.overlays[-(ind + 1)].offset = offset
            self.overlay_changed.emit(len(self.overlays) - (ind + 1))

    def reset_overlay_offsets(self):
        for ind, overlay in enumerate(self.overlays):
            overlay.offset = 0
            self.overlay_changed.emit(ind)

    def reset(self):
        for _ in range(len(self.overlays)):
            self.remove_overlay(0)



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import re
from typing import Optional
import numbers
import numpy as np


def _default_auto_level(data: np.ndarray) -> tuple[float, float]:
    """Compute colormap range from data through histogram

    :param data: Data from which to compute levels
    :returns: (lower limit, upper limit)
    """
    counts, bin_edges = np.histogram(data, bins=3000)
    left_edges = bin_edges[:-1]
    left_edges = left_edges[np.where(counts > 0)]
    counts = counts[np.where(counts > 0)]

    hist_y_cumsum = np.cumsum(counts)
    hist_y_sum = np.sum(counts)

    max_ind = np.where(hist_y_cumsum < (0.996 * hist_y_sum))
    min_level = np.mean(left_edges[:2])

    if len(max_ind[0]):
        max_level = left_edges[max_ind[0][-1]]
    else:
        max_level = 0.5 * np.max(left_edges)

    if len(left_edges[left_edges > 0]) > 0:
        min_level = max(min_level, np.nanmin(left_edges[left_edges > 0]))
    return min_level, max_level


def _minmax_auto_level(data: np.ndarray) -> tuple[float, float]:
    """Returns min/max of the data

    :param data: Data from which to compute levels
    :returns: (min, max)
    """
    return float(np.min(data)), float(np.max(data))


def _mean3std_auto_level(data: np.ndarray) -> tuple[float, float]:
    """Returns mean+/-3std clipped to min/max of the data

    :param data: Data from which to compute levels
    :returns: (lower limit, upper limit)
    """
    mean = np.mean(data, dtype=np.float64)
    std = np.std(data, dtype=np.float64)
    minimum, maximum = _minmax_auto_level(data)
    return max(mean - 3 * std, minimum), min(mean + 3 * std, maximum)


def _percentile_auto_level(
    data: np.ndarray, percentile: float = 1.0
) -> tuple[float, float]:
    """Returns data range corresponding to [precentile, 100-percentile]

    :param data: Data from which to compute levels
    :returns: (lower limit, upper limit)
    """
    if not 0.0 <= percentile < 50.0:
        raise ValueError("Percentiles must be in the range [0, 50[")
    lower, upper = np.percentile(data, [percentile, 100.0 - percentile])
    return float(lower), float(upper)


def _find_colums_with_same_value(data: np.ndarray) -> Optional[numbers.Number]:
    """Find value of columns with same value if any.

    :param data: 2D image array for which to compute the mask
    :returns: The value or None if not enough colunms with the same value
    """
    if data.ndim != 2:
        raise ValueError("2D array only are supported")

    # Find columns with same values
    equal_columns_mask = np.all(np.equal(data[0], data), axis=0)
    if not np.any(equal_columns_mask):
        return None

    equal_values, counts = np.unique(data[0, equal_columns_mask], return_counts=True)
    # At least 2 columns and not the whole image
    mask = np.logical_and(counts > 1, counts < data.shape[1])
    equal_values = equal_values[mask]
    counts = counts[mask]
    if len(equal_values) == 0:
        return None

    return equal_values[np.argmax(counts)]


def detector_gap_mask(data: np.ndarray) -> np.ndarray:
    """Probe detector gap value and returns mask of pixels not belonging to gaps.

    :param data: 2D image array for which to compute the mask
    :returns: Mask with True for valid pixels and False for mask
    """
    if data.ndim != 2:
        raise ValueError("2D array only are supported")

    # Find columns with same value
    value = _find_colums_with_same_value(data)
    if value is None:
        # Find rows with same value
        value = _find_colums_with_same_value(np.transpose(data))
    if value is None:
        return np.ones(data.shape, dtype=bool)

    return data != value


class AutoLevel:
    """Handle colormap range autoscale computation.

    This class stores settings: autoscale mode and whether or not to filter dummy value.

    Instances of this class are callable:
    >>> auto_level = AutoLevel()
    >>> range_ = auto_level.get_range(data)
    """

    _MODES = {
        "default": _default_auto_level,
        "minmax": _minmax_auto_level,
        "mean3std": _mean3std_auto_level,
    }
    _PERCENTILE_REGEXP = re.compile(r"(?P<value>\d+(\.\d*)?|\.\d+)percentile")

    def __init__(self):
        self.mode: str = "default"
        """Autoscale mode in 'default', 'minmax', 'mean3std', '%fpercentile'"""

        self.filter_dummy: bool = True
        """Whether or not to filter detector dummy values"""

    def get_range(self, data: Optional[np.ndarray]) -> Optional[tuple[float, float]]:
        """Returns colormap range from data for current settings

        :param data: Data from which to compute colormap range
        :returns: (min, max) or None
        """
        if data is None:
            return None

        if self.filter_dummy:
            data = data[detector_gap_mask(data)]

        filtered_data = data[np.isfinite(data)]
        if filtered_data.size == 0:
            return None

        func = self._MODES.get(self.mode, None)
        if func is not None:
            return func(filtered_data)
        match = self._PERCENTILE_REGEXP.match(self.mode)
        if match is not None:
            return _percentile_auto_level(filtered_data, percentile=float(match["value"]))
        raise ValueError(f"Unsupported mode: {self.mode}")


auto_level = AutoLevel()



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np


def convert_units(value, wavelength, previous_unit, new_unit):
    """
    Converts a value from a unit into a new unit
    :param value: value in old unit
    :param wavelength: in Angstrom
    :param previous_unit: possible values are '2th_deg', 'q_A^-1', 'd_A'
    :param new_unit: possible values are '2th_deg', 'q_A^-1', 'd_A'
    :return: new value or None if unit does not exist
    """
    if previous_unit == '2th_deg':
        tth = value
    elif previous_unit == 'q_A^-1':
        tth = np.arcsin(
            value * 1e10 * wavelength / (4 * np.pi)) * 360 / np.pi
    elif previous_unit == 'd_A':
        tth = 2 * np.arcsin(wavelength / (2 * value * 1e-10)) * 180 / np.pi
    else:
        tth = 0

    if new_unit == '2th_deg':
        res = tth
    elif new_unit == 'q_A^-1':
        res = 4 * np.pi * \
              np.sin(tth / 360 * np.pi) / \
              wavelength / 1e10
    elif new_unit == 'd_A':
        res = wavelength / (2 * np.sin(tth / 360 * np.pi)) * 1e10
    else:
        res = None
    return res


def supersample_image(img_data, factor):
    """
    Creates a supersampled array from img_data.
    :param img_data: image array
    :param factor: int - supersampling factor
    :return: supersampled image
    """
    if factor > 1:
        img_data_supersampled = np.zeros((img_data.shape[0] * factor,
                                          img_data.shape[1] * factor))
        for row in range(factor):
            for col in range(factor):
                img_data_supersampled[row::factor, col::factor] = img_data

        return img_data_supersampled
    else:
        return img_data


def trim_trailing_zeros(x, y):
    """
    Trims the trailing zeros of a x, y pattern
    :param x: x-values
    :param y: y-values
    :return: trimmed x, y values as tuple (x, y)
    """

    y_trim = np.trim_zeros(y, 'b')
    x_trim = x[:len(y_trim)]

    return x_trim, y_trim



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from functools import partial
import os

from qtpy import QtWidgets, QtCore, QtGui

from ...CustomWidgets import (
    FlatButton,
    DoubleSpinBoxAlignRight,
    VerticalSpacerItem,
    NoRectDelegate,
    ListTableWidget,
    HorizontalLine,
    DoubleMultiplySpinBoxAlignRight,
)

from .... import icons_path


class PhaseWidget(QtWidgets.QWidget):
    color_btn_clicked = QtCore.Signal(int, QtWidgets.QWidget)
    show_cb_state_changed = QtCore.Signal(int, bool)

    pressure_sb_value_changed = QtCore.Signal(int, float)
    temperature_sb_value_changed = QtCore.Signal(int, float)

    def __init__(self):
        super(PhaseWidget, self).__init__()

        self._layout = QtWidgets.QHBoxLayout()
        self._layout.setContentsMargins(5, 5, 5, 5)
        self._layout.setSpacing(5)

        self.add_btn = FlatButton()
        self.edit_btn = FlatButton()
        self.delete_btn = FlatButton()
        self.clear_btn = FlatButton()
        self.save_list_btn = QtWidgets.QPushButton("Save List")
        self.load_list_btn = QtWidgets.QPushButton("Load List")

        self.button_widget = QtWidgets.QWidget(self)
        self.button_widget.setObjectName("phase_control_button_widget")
        self._button_layout = QtWidgets.QVBoxLayout()
        self._button_layout.setContentsMargins(0, 0, 0, 0)
        self._button_layout.setSpacing(6)

        self._button_layout.addWidget(self.add_btn)
        self._button_layout.addWidget(self.edit_btn)
        self._button_layout.addWidget(HorizontalLine())
        self._button_layout.addWidget(self.delete_btn)
        self._button_layout.addWidget(HorizontalLine())
        self._button_layout.addWidget(self.clear_btn)
        self._button_layout.addSpacerItem(VerticalSpacerItem())
        self.button_widget.setLayout(self._button_layout)
        self._layout.addWidget(self.button_widget)

        self.parameter_widget = QtWidgets.QWidget()

        self._parameter_layout = QtWidgets.QVBoxLayout()
        self._parameter_layout.setContentsMargins(0, 0, 0, 0)
        self._parameter_layout.setSpacing(4)

        self.pressure_step_msb = DoubleMultiplySpinBoxAlignRight()
        self.temperature_step_msb = DoubleMultiplySpinBoxAlignRight()
        self.apply_to_all_cb = QtWidgets.QCheckBox("apply to all")

        self._parameter_layout.addWidget(QtWidgets.QLabel("P step"))
        self._parameter_layout.addWidget(self.pressure_step_msb)
        self._parameter_layout.addWidget(QtWidgets.QLabel("T Step"))
        self._parameter_layout.addWidget(self.temperature_step_msb)
        self._parameter_layout.addWidget(self.apply_to_all_cb)
        self._parameter_layout.addWidget(HorizontalLine())
        self._parameter_layout.addItem(VerticalSpacerItem())
        self._parameter_layout.addWidget(self.save_list_btn)
        self._parameter_layout.addWidget(self.load_list_btn)

        self.parameter_widget.setLayout(self._parameter_layout)

        self._body_layout = QtWidgets.QHBoxLayout()

        self.phase_tw = ListTableWidget(columns=5)
        self.phase_tw.setObjectName("phase_table_widget")
        self.phase_tw.setHorizontalHeaderLabels(["", "", "Name", "P (GPa)", "T (K)"])
        self.phase_tw.horizontalHeader().setVisible(True)
        self.phase_tw.horizontalHeader().setStretchLastSection(False)
        self.phase_tw.setColumnWidth(0, 20)
        self.phase_tw.setColumnWidth(1, 25)
        self.phase_tw.horizontalHeader().setSectionResizeMode(
            0, QtWidgets.QHeaderView.Fixed
        )
        self.phase_tw.horizontalHeader().setSectionResizeMode(
            1, QtWidgets.QHeaderView.Fixed
        )
        self.phase_tw.horizontalHeader().setSectionResizeMode(
            2, QtWidgets.QHeaderView.Stretch
        )
        self.phase_tw.horizontalHeader().setSectionResizeMode(
            3, QtWidgets.QHeaderView.ResizeToContents
        )
        self.phase_tw.horizontalHeader().setSectionResizeMode(
            4, QtWidgets.QHeaderView.ResizeToContents
        )
        self.phase_tw.setItemDelegate(NoRectDelegate())
        self._body_layout.addWidget(self.phase_tw, 10)
        self._body_layout.addWidget(self.parameter_widget, 0)

        self._layout.addLayout(self._body_layout)

        # label for alternative view:
        self.phase_header_btn = FlatButton("Phase")
        self.phase_header_btn.setObjectName("phase_header_btn")
        self.phase_header_btn.setEnabled(False)
        self.phase_header_btn.setVisible(False)
        self._main_layout = QtWidgets.QVBoxLayout()
        self._main_layout.setContentsMargins(0, 0, 0, 0)
        self._main_layout.setSpacing(0)
        self._header_layout = QtWidgets.QHBoxLayout()
        self._header_layout.setContentsMargins(0, 0, 0, 0)
        self._header_layout.setSpacing(0)
        self._header_layout.addWidget(self.phase_header_btn)
        self._header_layout.addStretch()
        self._main_layout.addLayout(self._header_layout)
        self._main_layout.addLayout(self._layout)
        self.setLayout(self._main_layout)
        self.style_widgets()
        self.add_tooltips()

        self.phase_show_cbs = []
        self.phase_color_btns = []
        self.pressure_sbs = []
        self.temperature_sbs = []

        self.show_parameter_in_pattern = True

    def style_widgets(self):
        icon_size = QtCore.QSize(17, 17)

        self.add_btn.setIcon(QtGui.QIcon(os.path.join(icons_path, "open.ico")))
        self.add_btn.setIconSize(icon_size)

        self.edit_btn.setIcon(QtGui.QIcon(os.path.join(icons_path, "edit.png")))
        self.edit_btn.setIconSize(QtCore.QSize(14, 14))

        self.delete_btn.setIcon(QtGui.QIcon(os.path.join(icons_path, "delete.png")))
        self.delete_btn.setIconSize(QtCore.QSize(12, 14))

        self.clear_btn.setIcon(QtGui.QIcon(os.path.join(icons_path, "reset_dark.ico")))
        self.clear_btn.setIconSize(icon_size)

        def modify_btn_to_icon_size(btn):
            button_height = 25
            button_width = 25
            btn.setMinimumHeight(button_height)
            btn.setMaximumHeight(button_height)
            btn.setMinimumWidth(button_width)
            btn.setMaximumWidth(button_width)

        modify_btn_to_icon_size(self.add_btn)
        modify_btn_to_icon_size(self.delete_btn)
        modify_btn_to_icon_size(self.clear_btn)
        modify_btn_to_icon_size(self.edit_btn)

        self.phase_tw.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.MinimumExpanding
        )
        self.parameter_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum
        )

        step_txt_width = 80

        self.pressure_step_msb.setMinimumWidth(step_txt_width)
        self.pressure_step_msb.setMaximumWidth(step_txt_width)
        self.temperature_step_msb.setMinimumWidth(step_txt_width)
        self.temperature_step_msb.setMaximumWidth(step_txt_width)

        self.temperature_step_msb.setMaximum(1000.0)
        self.temperature_step_msb.setMinimum(1.0)
        self.temperature_step_msb.setValue(100.0)

        self.pressure_step_msb.setValue(1)

        self.phase_header_btn.setStyleSheet("border-radius: 0px")

        self.apply_to_all_cb.setChecked(True)

    def add_tooltips(self):
        self.add_btn.setToolTip("Loads Phase(s) from jcpds or cif file(s)")
        self.edit_btn.setToolTip("Edit selected Phase")
        self.delete_btn.setToolTip("Removes currently selected phase")
        self.clear_btn.setToolTip("Removes all phases")
        self.apply_to_all_cb.setToolTip(
            "Whether individual changes in P or T\nare applied to all other phases"
        )
        self.pressure_step_msb.setToolTip("Sets the step for the pressure spinboxes")
        self.temperature_step_msb.setToolTip(
            "Sets the step for the temperature spinboxes"
        )

    # ###############################################################################################
    # Now comes all the phase tw stuff
    ################################################################################################

    def add_phase(self, name, color):
        current_rows = self.phase_tw.rowCount()
        self.phase_tw.setRowCount(current_rows + 1)
        self.phase_tw.blockSignals(True)

        show_cb = QtWidgets.QCheckBox()
        show_cb.setChecked(True)
        show_cb.stateChanged.connect(partial(self.phase_show_cb_changed, show_cb))
        show_cb.setStyleSheet("background-color: transparent")
        self.phase_tw.setCellWidget(current_rows, 0, show_cb)
        self.phase_show_cbs.append(show_cb)

        color_button = FlatButton()
        color_button.setStyleSheet(f"background-color: {color}; margin: 2px;")
        color_button.clicked.connect(partial(self.phase_color_btn_click, color_button))
        self.phase_tw.setCellWidget(current_rows, 1, color_button)
        self.phase_color_btns.append(color_button)

        name_item = QtWidgets.QTableWidgetItem(name)
        name_item.setFlags(name_item.flags() & ~QtCore.Qt.ItemIsEditable)
        name_item.setTextAlignment(int(QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter))
        self.phase_tw.setItem(current_rows, 2, name_item)

        pressure_sb = DoubleSpinBoxAlignRight()
        pressure_sb.setFixedWidth(70)
        pressure_sb.setMinimum(-9999999)
        pressure_sb.setMaximum(9999999)
        pressure_sb.setValue(0)
        pressure_sb.setSingleStep(self.pressure_step_msb.value())
        pressure_sb.valueChanged.connect(
            partial(self.pressure_sb_callback, pressure_sb)
        )
        self.phase_tw.setCellWidget(current_rows, 3, pressure_sb)
        self.pressure_sbs.append(pressure_sb)

        temperature_sb = DoubleSpinBoxAlignRight()
        temperature_sb.setFixedWidth(80)
        temperature_sb.setMinimum(-9999999)
        temperature_sb.setMaximum(9999999)
        temperature_sb.setValue(300)
        temperature_sb.setSingleStep(self.temperature_step_msb.value())
        temperature_sb.valueChanged.connect(
            partial(self.temperature_sb_callback, temperature_sb)
        )
        self.phase_tw.setCellWidget(current_rows, 4, temperature_sb)
        self.temperature_sbs.append(temperature_sb)

        self.phase_tw.setRowHeight(current_rows, 25)
        self.select_phase(current_rows)
        self.phase_tw.blockSignals(False)

    def update_phase_tw_column_sizes(self):
        self.phase_tw.setColumnWidth(0, 20)
        self.phase_tw.setColumnWidth(1, 25)
        self.phase_tw.horizontalHeader().setSectionResizeMode(
            3, QtWidgets.QHeaderView.ResizeToContents
        )
        self.phase_tw.horizontalHeader().setSectionResizeMode(
            4, QtWidgets.QHeaderView.ResizeToContents
        )

    def select_phase(self, ind):
        self.phase_tw.selectRow(ind)

    def get_selected_phase_row(self):
        selected = self.phase_tw.selectionModel().selectedRows()
        try:
            row = selected[0].row()
        except IndexError:
            row = -1
        return row

    def get_phase(self):
        pass

    def del_phase(self, ind):
        self.phase_tw.blockSignals(True)
        self.phase_tw.removeRow(ind)
        self.phase_tw.blockSignals(False)
        del self.phase_show_cbs[ind]
        del self.phase_color_btns[ind]
        del self.temperature_sbs[ind]
        del self.pressure_sbs[ind]

        if self.phase_tw.rowCount() > ind:
            self.select_phase(ind)
        else:
            self.select_phase(self.phase_tw.rowCount() - 1)

    def rename_phase(self, ind, name):
        name_item = self.phase_tw.item(ind, 2)
        name_item.setText(name)

    def set_phase_temperature(self, ind, temperature):
        pass
        self.temperature_sbs[ind].blockSignals(True)
        self.temperature_sbs[ind].setValue(temperature)
        self.temperature_sbs[ind].blockSignals(False)

    def get_phase_temperature(self, ind):
        return self.temperature_sbs[ind].value()

    def set_phase_pressure(self, ind, pressure):
        self.pressure_sbs[ind].blockSignals(True)
        self.pressure_sbs[ind].setValue(pressure)
        self.pressure_sbs[ind].blockSignals(False)

    def get_phase_pressure(self, ind):
        return self.pressure_sbs[ind].value()

    def phase_color_btn_click(self, button):
        self.color_btn_clicked.emit(self.phase_color_btns.index(button), button)

    def phase_show_cb_changed(self, checkbox):
        self.show_cb_state_changed.emit(
            self.phase_show_cbs.index(checkbox), checkbox.isChecked()
        )

    def phase_show_cb_set_checked(self, ind, state):
        checkbox = self.phase_show_cbs[ind]
        checkbox.setChecked(state)

    def phase_show_cb_is_checked(self, ind):
        checkbox = self.phase_show_cbs[ind]
        return checkbox.isChecked()

    def pressure_sb_callback(self, pressure_sb):
        self.pressure_sb_value_changed.emit(
            self.pressure_sbs.index(pressure_sb), pressure_sb.value()
        )

    def temperature_sb_callback(self, temperature_sb):
        self.temperature_sb_value_changed.emit(
            self.temperature_sbs.index(temperature_sb), temperature_sb.value()
        )



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ....model.util.HelperModule import get_partial_index

# imports for type hinting in PyCharm -- DO NOT DELETE
from ....model.DioptasModel import DioptasModel
from ....widgets.integration import IntegrationWidget
from ....widgets.plot_widgets.ImgWidget import IntegrationImgWidget


class PhaseInCakeController(object):
    """
    PhaseInCakeController handles all the interaction between the phase controls and the plotted lines in the cake view.
    """

    def __init__(self, integration_widget: IntegrationWidget, dioptas_model: DioptasModel):
        """
        :param integration_widget: Reference to an IntegrationWidget
        :param dioptas_model: reference to DioptasModel object
        """
        self.model = dioptas_model
        self.phase_model = self.model.phase_model
        self.integration_widget = integration_widget
        self.cake_view_widget = integration_widget.integration_image_widget.cake_view  # type: IntegrationImgWidget

        self.connect()

    def connect(self):
        self.phase_model.phase_added.connect(self.add_phase_plot)
        self.model.phase_model.phase_removed.connect(self.cake_view_widget.del_cake_phase)

        self.phase_model.phase_changed.connect(self.update_phase_lines)
        self.phase_model.phase_changed.connect(self.update_phase_color)
        self.phase_model.phase_changed.connect(self.update_phase_visible)

        self.phase_model.reflection_added.connect(self.reflection_added)
        self.phase_model.reflection_deleted.connect(self.reflection_deleted)

    def get_phase_position_and_intensities(self, ind, clip=True):
        """
        Obtains the positions and intensities for lines of a phase with an index ind within the cake view.

        No clipping is used for the first call to add the CakePhasePlot to the ImgWidget. Subsequent calls are used with
        clipping. Thus, only lines within the cake_tth are returned. The visibility of each line is then estimated in
        the ImgWidget based on the length of the clipped and not clipped lists.

        :param ind: the index of the phase
        :param clip: whether or not the lists should be clipped. Clipped means that lines which have positions larger
                     than the
        :return: line_positions, line_intensities
        """
        if self.model.cake_tth is None:
            cake_tth = self.model.calibration_model.tth
        else:
            cake_tth = self.model.cake_tth
        reflections_tth = self.phase_model.get_phase_line_positions(ind, '2th_deg',
                                                                    self.model.calibration_model.wavelength * 1e10)
        reflections_intensities = [reflex[1] for reflex in self.phase_model.reflections[ind]]

        cake_line_positions = []
        cake_line_intensities = []

        for ind, tth in enumerate(reflections_tth):
            pos_ind = get_partial_index(cake_tth, tth)
            if pos_ind is not None:
                cake_line_positions.append(pos_ind + 0.5)
                cake_line_intensities.append(reflections_intensities[ind])
            elif clip is False:
                cake_line_positions.append(0)
                cake_line_intensities.append(reflections_intensities[ind])

        return cake_line_positions, cake_line_intensities

    def add_phase_plot(self):
        cake_line_positions, cake_line_intensities = self.get_phase_position_and_intensities(-1, False)

        self.cake_view_widget.add_cake_phase(cake_line_positions, cake_line_intensities,
                                             self.phase_model.phase_colors[-1])

    def update_phase_lines(self, ind):
        cake_line_positions, cake_line_intensities = self.get_phase_position_and_intensities(ind)
        self.cake_view_widget.update_phase_intensities(ind, cake_line_positions, cake_line_intensities)

    def update_phase_color(self, ind):
        self.cake_view_widget.set_cake_phase_color(ind, self.model.phase_model.phase_colors[ind])

    def update_phase_visible(self, ind):
        if self.phase_model.phase_visible[ind] and self.integration_widget.img_mode == 'Cake' and \
                self.integration_widget.img_phases_btn.isChecked():
            self.cake_view_widget.show_cake_phase(ind)
        else:
            self.cake_view_widget.hide_cake_phase(ind)

    def reflection_added(self, ind):
        self.cake_view_widget.phases[ind].add_line()

    def reflection_deleted(self, phase_ind, reflection_ind):
        self.cake_view_widget.phases[phase_ind].delete_line(reflection_ind)



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from qtpy import QtWidgets, QtCore

from ...CustomWidgets import (
    MenuTabWidget,
    NumberTextField,
    LabelAlignRight,
    SpinBoxAlignRight,
    CheckableFlatButton,
    DoubleSpinBoxAlignRight,
    HorizontalSpacerItem,
    DoubleMultiplySpinBoxAlignRight,
    SaveIconButton,
)


class BackgroundWidget(QtWidgets.QWidget):
    def __init__(self):
        super(BackgroundWidget, self).__init__()

        self._layout = QtWidgets.QHBoxLayout()
        self._layout.setContentsMargins(0, 5, 0, 0)
        self._layout.setSpacing(5)
        self.setLayout(self._layout)

        self.create_image_background_widgets()
        self.create_image_background_layout()

        self.create_pattern_background_widgets()
        self.create_pattern_background_layout()

        self._tab_widget = MenuTabWidget()
        self._tab_widget.add_tab("Image Bkg", self.image_background_gb)
        self._tab_widget.add_tab("Pattern Bkg", self.pattern_background_gb)
        self._layout.addWidget(self._tab_widget)

        self.style_widgets()

    def create_image_background_widgets(self):
        self.image_background_gb = QtWidgets.QGroupBox("Image Background", self)
        self._image_background_gb_layout = QtWidgets.QGridLayout(
            self.image_background_gb
        )
        self._image_background_gb_layout.setContentsMargins(5, 8, 5, 7)
        self._image_background_gb_layout.setSpacing(5)

        self.load_image_btn = QtWidgets.QPushButton("Load")
        self.filename_lbl = QtWidgets.QLabel("None")
        self.remove_image_btn = QtWidgets.QPushButton("Remove")
        self.scale_sb = DoubleSpinBoxAlignRight()
        self.offset_sb = DoubleSpinBoxAlignRight()

        self.scale_step_msb = DoubleMultiplySpinBoxAlignRight()
        self.offset_step_msb = DoubleMultiplySpinBoxAlignRight()

    def create_image_background_layout(self):
        self._image_background_gb_layout.addWidget(self.load_image_btn, 0, 0)
        self._image_background_gb_layout.addWidget(self.remove_image_btn, 1, 0)
        self._image_background_gb_layout.addWidget(self.filename_lbl, 0, 1, 1, 8)
        self._image_background_gb_layout.addWidget(LabelAlignRight("Scale:"), 1, 1)
        self._image_background_gb_layout.addWidget(self.scale_sb, 1, 2)
        self._image_background_gb_layout.addWidget(self.scale_step_msb, 1, 3)
        self._image_background_gb_layout.addItem(HorizontalSpacerItem(), 1, 4)
        self._image_background_gb_layout.addWidget(LabelAlignRight("Offset:"), 2, 1)
        self._image_background_gb_layout.addWidget(self.offset_sb, 2, 2)
        self._image_background_gb_layout.addWidget(self.offset_step_msb, 2, 3)
        self._image_background_gb_layout.setRowStretch(0, 0)
        self._image_background_gb_layout.setRowStretch(1, 0)
        self._image_background_gb_layout.setRowStretch(2, 0)
        self._image_background_gb_layout.setRowStretch(3, 1)

        self.image_background_gb.setLayout(self._image_background_gb_layout)

    def create_pattern_background_widgets(self):
        self.pattern_background_gb = QtWidgets.QGroupBox("Pattern Background")
        self._pattern_bkg_layout = QtWidgets.QGridLayout()
        self._pattern_bkg_layout.setContentsMargins(5, 8, 5, 7)
        self._pattern_bkg_layout.setSpacing(5)

        self.smooth_with_sb = DoubleSpinBoxAlignRight()
        self.iterations_sb = SpinBoxAlignRight()
        self.poly_order_sb = SpinBoxAlignRight()
        self.x_range_min_txt = NumberTextField("0")
        self.x_range_max_txt = NumberTextField("50")
        self.inspect_btn = CheckableFlatButton("Inspect")
        self.save_btn = SaveIconButton()
        self.as_overlay = QtWidgets.QPushButton("As Overlay")

    def create_pattern_background_layout(self):
        self._pattern_bkg_layout.addWidget(LabelAlignRight("Smooth Width:"), 0, 0)
        self._pattern_bkg_layout.addWidget(self.smooth_with_sb, 0, 1)
        self._pattern_bkg_layout.addWidget(LabelAlignRight("Iterations:"), 0, 2)
        self._pattern_bkg_layout.addWidget(self.iterations_sb, 0, 3)
        self._pattern_bkg_layout.addWidget(LabelAlignRight("Order:"), 1, 0)
        self._pattern_bkg_layout.addWidget(self.poly_order_sb, 1, 1)

        self._pattern_bkg_layout.addWidget(LabelAlignRight("X-Range:"), 2, 0)
        self._range_layout = QtWidgets.QHBoxLayout()
        self._range_layout.addWidget(self.x_range_min_txt)
        self._range_layout.addWidget(QtWidgets.QLabel("-"))
        self._range_layout.addWidget(self.x_range_max_txt)
        self._range_layout.addItem(HorizontalSpacerItem())
        self._pattern_bkg_layout.addLayout(self._range_layout, 2, 1, 1, 3)
        self._button_layout = QtWidgets.QHBoxLayout()
        self._button_layout.addStretch()
        self._button_layout.addWidget(self.inspect_btn)
        self._button_layout.addWidget(self.as_overlay)
        self._button_layout.addWidget(self.save_btn)
        self._pattern_bkg_layout.addLayout(self._button_layout, 3, 0, 1, 4)

        self._pattern_bkg_layout.setRowStretch(0, 0)
        self._pattern_bkg_layout.setRowStretch(1, 0)
        self._pattern_bkg_layout.setRowStretch(2, 0)
        self._pattern_bkg_layout.setRowStretch(5, 1)
        self._pattern_bkg_layout.setColumnStretch(0, 0)
        self._pattern_bkg_layout.setColumnStretch(1, 0)
        self._pattern_bkg_layout.setColumnStretch(2, 0)
        self._pattern_bkg_layout.setColumnStretch(3, 0)
        self._pattern_bkg_layout.setColumnStretch(4, 1)

        self.pattern_background_gb.setLayout(self._pattern_bkg_layout)

    def style_widgets(self):
        self.style_image_background_widgets()
        self.style_pattern_background_widgets()

    def style_image_background_widgets(self):
        self.scale_sb.setMinimum(-9999999)
        self.scale_sb.setMaximum(9999999)
        self.scale_sb.setValue(1)
        self.scale_sb.setSingleStep(0.01)

        self.pattern_background_gb.setCheckable(True)
        self.pattern_background_gb.setChecked(False)

        self.scale_step_msb.setMaximum(10.0)
        self.scale_step_msb.setMinimum(0.01)
        self.scale_step_msb.setValue(0.01)

        self.offset_step_msb.setMaximum(100000.0)
        self.offset_step_msb.setMinimum(0.01)
        self.offset_step_msb.setValue(100.0)

        self.offset_sb.setMaximum(999999998)
        self.offset_sb.setMinimum(-99999999)

        self.setStyleSheet(
            """
            QSpinBox, QDoubleSpinBox, QLineEdit {
                width: 60px;
                min-width: 30px;
                max-width: 80px;
            } 
        """
        )

    def style_pattern_background_widgets(self):
        self.smooth_with_sb.setValue(0.100)
        self.smooth_with_sb.setMinimum(0)
        self.smooth_with_sb.setMaximum(10000000)
        self.smooth_with_sb.setSingleStep(0.005)
        self.smooth_with_sb.setDecimals(3)
        self.smooth_with_sb.setMaximumWidth(100)

        self.iterations_sb.setMaximum(99999)
        self.iterations_sb.setMinimum(1)
        self.iterations_sb.setValue(150)
        self.poly_order_sb.setMaximum(999999)
        self.poly_order_sb.setMinimum(1)
        self.poly_order_sb.setValue(50)
        self.poly_order_sb.setToolTip("Set the Polynomial order")

        self.save_btn.setToolTip("Save generated background pattern")
        self.save_btn.setIconSize(QtCore.QSize(13, 13))
        self.save_btn.setMaximumWidth(25)

    def get_bkg_pattern_parameters(self):
        smooth_width = float(self.smooth_with_sb.value())
        iterations = int(self.iterations_sb.value())
        polynomial_order = int(self.poly_order_sb.value())
        return smooth_width, iterations, polynomial_order

    def set_bkg_pattern_parameters(self, bkg_pattern_parameters):
        self.smooth_with_sb.blockSignals(True)
        self.iterations_sb.blockSignals(True)
        self.poly_order_sb.blockSignals(True)

        self.smooth_with_sb.setValue(bkg_pattern_parameters[0])
        self.iterations_sb.setValue(bkg_pattern_parameters[1])
        self.poly_order_sb.setValue(bkg_pattern_parameters[2])

        self.smooth_with_sb.blockSignals(False)
        self.iterations_sb.blockSignals(False)
        self.poly_order_sb.blockSignals(False)

    def get_bkg_pattern_roi(self):
        x_min = float(str(self.x_range_min_txt.text()))
        x_max = float(str(self.x_range_max_txt.text()))
        return x_min, x_max

    def set_bkg_pattern_roi(self, roi):
        self.x_range_min_txt.blockSignals(True)
        self.x_range_max_txt.blockSignals(True)

        self.x_range_min_txt.setText("{:.3f}".format(roi[0]))
        self.x_range_max_txt.setText("{:.3f}".format(roi[1]))

        self.x_range_min_txt.blockSignals(False)
        self.x_range_max_txt.blockSignals(False)



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import copy

import numpy as np
from PIL import Image
import h5py

import fabio

from .util import Signal
from dioptas.model.loader.spe import SpeFile
from .util.NewFileWatcher import NewFileInDirectoryWatcher
from .util.HelperModule import rotate_matrix_p90, rotate_matrix_m90, FileNameIterator
from .util.ImgCorrection import ImgCorrectionManager, ImgCorrectionInterface, TransferFunctionCorrection
from dioptas.model.loader.LambdaLoader import LambdaImage
from dioptas.model.loader.KaraboLoader import KaraboFile
from dioptas.model.loader.hdf5Loader import Hdf5Image
from dioptas.model.loader.FabioLoader import FabioLoader

logger = logging.getLogger(__name__)


class ImgModel(object):
    """
    Main Image handling class. Supports several features:
        - loading image files in any format using fabio
        - iterating through files either by file number or time of creation
        - image transformations like rotating and flipping
        - setting a background image
        - setting an absorption correction (img_data is divided by this)

    In order to subscribe to changes of the data in the ImgModel, please use the img_changed QtSignal.
    The Signal will be called every time the img_data has changed.
    """

    def __init__(self):
        super(ImgModel, self).__init__()
        self.filename = ''
        self.img_transformations = []

        self.file_iteration_mode = 'number'
        self.file_name_iterator = FileNameIterator()

        self.series_pos = 1
        self.series_max = 1
        self.selected_source = None

        self._img_data = None
        self._img_data_background_subtracted = None
        self._img_data_absorption_corrected = None
        self._img_data_background_subtracted_absorption_corrected = None

        self.background_filename = ''
        self._background_data = None
        self._background_scaling = 1
        self._background_offset = 0

        self._factor = 1

        self.transfer_correction = TransferFunctionCorrection()

        # anything that gets loaded from an image file and needs to be reset if a file without these attributes is
        # loaded 2D array containing the current image
        self.loadable_data = [
            {"name": "img_data", "default": np.zeros((2048, 2048)), "attribute": "_img_data"},
            {"name": "file_info", "default": "", "attribute": "file_info"},
            {"name": "motors_info", "default": {}, "attribute": "motors_info"},
            {"name": "img_data_fabio", "default": None, "attribute": "_img_data_fabio"},

            # current position in the loaded series of images, starting at 1
            {"name": "series_pos", "default": 1, "attribute": "series_pos"},

            # maximum position/number of images in the loaded series, starting at 1
            {"name": "series_max", "default": 1, "attribute": "series_max"},

            # function to get an image in the current series. A function assigned to this attribute should take
            # a single parameter pos (position in the series starting at 0) and return a 2d array with the image data
            {"name": "series_get_image", "default": None, "attribute": "series_get_image"},

            # list of sources for different image series within 1 file. This is used by an HDF5 file with several
            # datasets
            {"name": "sources", "default": None, "attribute": "sources"},

            # a function to select a source:
            {"name": "select_source", "default": None, "attribute": "_select_source"}
        ]

        # set the loadable attributes to their defaults
        self.set_loadable_attributes({})

        self._img_corrections = ImgCorrectionManager()

        # setting up autoprocess
        self._autoprocess = False
        self._directory_watcher = NewFileInDirectoryWatcher(
            file_types=['img', 'sfrm', 'dm3', 'edf', 'xml',
                        'cbf', 'kccd', 'msk', 'spr', 'tif',
                        'mccd', 'mar3450', 'pnm', 'spe']
        )
        self._directory_watcher.file_added.connect(self.load)

        # define the signals
        self.img_changed = Signal()
        self.autoprocess_changed = Signal()
        self.transformations_changed = Signal()
        self.corrections_removed = Signal()

    def load(self, filename, pos=0):
        """
        Loads an image file in any format known by fabIO, PIL or HDF5. Automatically performs all previous img
        transformations, recalculates background subtracted and absorption corrected image data.
        The img_changed signal will be emitted after the process.
        :param filename: path of the image file to be loaded
        :param pos: position of image in the image file to be loaded
        """
        filename = str(filename)  # since it could also be QString
        logger.info("Loading {0}.".format(filename))
        self.filename = filename

        image_file_data = self.get_image_data(filename, pos)
        self.set_loadable_attributes(image_file_data)

        self.file_name_iterator.update_filename(filename)
        self._directory_watcher.path = os.path.dirname(str(filename))

        self._perform_img_transformations()
        self._calculate_img_data()
        self.series_pos = pos + 1

        self.img_changed.emit()

    def get_image_data(self, filename, pos=0):
        """
        Tries to load the given file using different image loader libraries and returns a dictionary containing all
        retrieved file data.
        :param filename: string containing a path to an image file
        :param pos: position of image in the image file to be loaded
        :return: dictionary containing all retrieved file information. Look at "loadable data" for possible key names.
                 Present key names depend on applied image loader
        """
        img_loaders = [self.load_PIL, self.load_spe, self.load_fabio, self.load_lambda, self.load_karabo,
                       self.load_hdf5]

        for loader in img_loaders:
            data = loader(filename, pos)
            if data:
                return data
        else:
            raise IOError("No handler found for given image with filename: " + filename)

    def set_loadable_attributes(self, loaded_data):
        """
        Sets all attributes that change with the loading of an image to either their defaults or a given value.
        This assures that no leftover data will be kept when it is not overwritten by the new image.
        :param loaded_data: dictionary containing values to be loaded into the attributes corresponding to their keys.
                            Possible key names and attribute names they will be loaded to are specified in
                            "loadable_data"
        """
        for attribute in self.loadable_data:
            if attribute["name"] in loaded_data:
                self.__setattr__(attribute["attribute"], loaded_data[attribute["name"]])
            else:
                self.__setattr__(attribute["attribute"], copy.copy(attribute["default"]))

    def load_PIL(self, filename, *args):
        """
        Loads an image using the PIL library. Also returns file and motor info if present
        :param filename: path to the image file to be loaded
        :return: dictionary with image_data and file_info and motors_info if present. None if unsuccessful
        """
        data = {}
        try:
            im = Image.open(filename)
            if np.prod(im.size) <= 1:
                im.close()
                return False
            data["img_data"] = np.array(im)[::-1]
            try:
                data["file_info"] = self._get_file_info(im)
                data["motors_info"] = self._get_motors_info(im)
            except AttributeError:
                pass
            im.close()
            return data

        except IOError:
            return None

    def load_spe(self, filename, *args):
        """
        Loads an image using the builtin spe library.
        :param filename: path to the image file to be loaded
        :return: dictionary with image_data, None if unsuccessful
        """
        if os.path.splitext(filename)[1].lower() == '.spe':
            spe = SpeFile(filename)
            return {"img_data": spe.img}
        else:
            return None

    def load_fabio(self, filename, frame_index=0):
        """
        Loads an image using the fabio library.
        :param filename: path to the image file to be loaded
        :param frame_index: frame index of the image file to be loaded inside of multi-frame file
        :return: dictionary with image_data and image_data_fabio, None if unsuccessful
        """
        try:
            self.loader = FabioLoader(filename)
            return {
                "img_data_fabio": self.loader.fabio_image,
                "img_data": self.loader.get_image(frame_index),
                "series_max": self.loader.series_max,
                "series_get_image": self.loader.get_image
            }
        except (IOError, fabio.fabioutils.NotGoodReader):
            return None

    def load_lambda(self, filename, frame_index=0):
        """
        loads an image made by a lambda detector using the builtin lambda library.
        :param filename: path to the image file to be loaded
        :param frame_index: frame index of the image file to be loaded inside of multi-frame file
        :return: dictionary with img_data, series_max and series_get_image, None if unsuccessful
        """
        try:
            lambda_im = LambdaImage(filename)
        except IOError:
            return None

        if frame_index >= lambda_im.series_max:
            return None
        return {"img_data": lambda_im.get_image(frame_index),
                "series_max": lambda_im.series_max,
                "series_get_image": lambda_im.get_image}

    def load_karabo(self, filename, frame_index=0):
        """
        Loads an Imageseries created from within the karabo-framework at XFEL.
        :param filename: path to the *.h5 karabo file
        :param frame_index: position of image in the image file to be loaded
        :return: dictionary with img_data of the first train_id, series_start, series_max and series_get_image,
                 None if unsuccessful
        """
        try:
            karabo_file = KaraboFile(filename)
        except IOError:
            return None
        if frame_index >= karabo_file.series_max:
            return None
        return {"img_data": karabo_file.get_image(frame_index),
                "series_max": karabo_file.series_max,
                "series_get_image": karabo_file.get_image}

    def load_hdf5(self, filename, frame_index=0):
        """
        Loads an ESRF hdf5 file
        :param filename: filename with path to *.h5 ESRF file
        :param frame_index: frame index for multi-image file
        :return: dictionary with img_data of the first image in the first source, dataset_list, series_max, and
                 series_get_image
        """

        hdf5_image = Hdf5Image(filename)
        self.loader = hdf5_image
        self.selected_source = hdf5_image.image_sources[0]

        return {"img_data": hdf5_image.get_image(frame_index),
                "series_max": hdf5_image.series_max,
                "series_get_image": hdf5_image.get_image,
                "sources": hdf5_image.image_sources,
                "select_source": hdf5_image.select_source
                }

    def select_source(self, source):
        """
        Selects a source from the available sources and loads updates the current image in the model.
        :param source: string for source (check sources for available strings for the corresponding file)
        """
        self._select_source(source)
        self.selected_source = source
        self.series_max = self.loader.series_max
        self.series_pos = min(self.series_pos, self.series_max)
        self._img_data = self.series_get_image(self.series_pos - 1)

        self._perform_img_transformations()
        self._calculate_img_data()

        self.img_changed.emit()

    def save(self, filename):
        """
        Saves the current file as another image file, the raw data is used for saving.
        :param filename: name of the saved file, extensions defines the format, please see fabio library for reference
        """
        try:
            self._img_data_fabio.save(filename)
        except AttributeError:
            im_array = np.int32(np.copy(np.flipud(self._img_data)))
            im = Image.fromarray(im_array)
            im.save(filename)

    def load_background(self, filename):
        """
        Loads an image file as background in any format known by fabIO. Automatically performs all previous img
        transformations, recalculates background subtracted and absorption corrected image data.
        The img_changed signal will be emitted after the process.
        :param filename: path of the image file to be loaded
        """
        self.background_filename = filename

        self._background_data = self.get_image_data(filename)["img_data"]

        self._perform_background_transformations()

        if self._background_data.shape != self._img_data.shape:
            self._background_data = None
            self._calculate_img_data()
            self.img_changed.emit()
            raise BackgroundDimensionWrongException()

        self._calculate_img_data()
        self.img_changed.emit()

    def add(self, filename):
        """
        Adds an image file in any format known by fabIO. Automatically performs all previous img transformations and
        recalculates background subtracted and absorption corrected image data.
        The img_changed signal will be emitted after the process.
        :param filename: path of the image file to be loaded
        """
        filename = str(filename)  # since it could also be QString

        img_data = self.get_image_data(filename)["img_data"]

        for transformation in self.img_transformations:
            img_data = transformation(img_data)

        if not self._img_data.shape == img_data.shape:
            return

        logger.info("Adding {0}.".format(filename))

        if self._img_data.dtype == np.uint16:  # if dtype is only uint16 we will convert to 32 bit, so that more
            # additions are possible
            self._img_data = self._img_data.astype(np.uint32)

        self._img_data += img_data

        self._calculate_img_data()
        self.img_changed.emit()

    def _image_and_background_shape_equal(self):
        """
        Tests if the original image and original background image have the same shape
        :return: Boolean
        """
        if self._background_data is None:
            return True
        if self._background_data.shape == self._img_data.shape:
            return True
        return False

    def _reset_background(self):
        """
        Resets the background data to None
        """
        self.background_filename = ''
        self._background_data = None
        self._background_data_fabio = None
        self._calculate_img_data()

    def reset_background(self):
        self._reset_background()
        self.img_changed.emit()

    def has_background(self):
        return self._background_data is not None

    @property
    def background_data(self):
        return self._background_data

    @property
    def untransformed_background_data(self):
        self._reset_background_transformations()
        background_data = np.copy(self.background_data)
        self._perform_background_transformations()
        return background_data

    @background_data.setter
    def background_data(self, new_data):
        self._background_data = new_data
        self._calculate_img_data()
        self.img_changed.emit()

    @property
    def background_scaling(self):
        return self._background_scaling

    @background_scaling.setter
    def background_scaling(self, new_value):
        self._background_scaling = new_value
        self._calculate_img_data()
        self.img_changed.emit()

    @property
    def background_offset(self):
        return self._background_offset

    @background_offset.setter
    def background_offset(self, new_value):
        self._background_offset = new_value
        self._calculate_img_data()
        self.img_changed.emit()

    def load_series_img(self, pos):
        """
        Takes a position in  the series to load, sanitizes it and puts the result from the function assigned to
        series_get_image into _img_data. series_get_image gets called with a position starting from 0, all other series
        pos values start at one as shown to the user.
        :param pos: Image position in the series to load, starting at 1
        """
        pos = min(max(pos, 1), self.series_max)
        if self.series_pos == pos:
            return

        self.series_pos = pos
        self._img_data = self.series_get_image(pos - 1)

        self._perform_img_transformations()
        self._calculate_img_data()

        self.img_changed.emit()

    def load_next_file(self, step=1, pos=None):
        """
        Loads the next file based on the current iteration mode and the step you specify.
        :param pos:
        :param step: Defining how much you want to increment the file number. (default=1)
        """
        next_file_name = self.file_name_iterator.get_next_filename(mode=self.file_iteration_mode, step=step, pos=pos)
        if next_file_name is not None:
            self.load(next_file_name)

    def load_previous_file(self, step=1, pos=None):
        """
        Loads the previous file based on the current iteration mode and the step specified
        :param pos:
        :param step: Defining how much you want to decrement the file number. (default=1)
        """
        previous_file_name = self.file_name_iterator.get_previous_filename(mode=self.file_iteration_mode,
                                                                           step=step, pos=pos)
        if previous_file_name is not None:
            self.load(previous_file_name)

    def load_next_folder(self, mec_mode=False):
        """
        Loads a file with the current filename in the next folder, whereby the folder has to be iteratable by numbers.
        :param mec_mode:    Boolean which enables specific mode for MEC beamline at SLAC, where the folders and the
                            files change their during increment. (default = False)

        """
        next_file_name = self.file_name_iterator.get_next_folder(mec_mode=mec_mode)
        if next_file_name is not None:
            self.load(next_file_name)

    def load_previous_folder(self, mec_mode=False):
        """
        Loads a file with the current filename in the previous folder, whereby the folder has to be iteratable by
        numbers.
        :param mec_mode:    Boolean which enables specific mode for MEC beamline at SLAC, where the folders and the
                            files change their during increment. (default = False)
        """

        next_previous_name = self.file_name_iterator.get_previous_folder(mec_mode=mec_mode)
        if next_previous_name is not None:
            self.load(next_previous_name)

    def set_file_iteration_mode(self, mode):
        """
        Sets the file iteration mode for the load_next_file and load_previous_file functions. Possible modes:
            * 'number' will increment or decrement based on numbers in the filename.
            * 'time' will increment or decrement based on creation time for the files.
        """
        if mode == 'number':
            self.file_iteration_mode = 'number'
            self.file_name_iterator.create_timed_file_list = False
        elif mode == 'time':
            self.file_iteration_mode = 'time'
            self.file_name_iterator.create_timed_file_list = True
            self.file_name_iterator.update_filename(self.filename)

    def _calculate_img_data(self):
        """
        Calculates compound img_data based on the state of the object. This function is used internally to not compute
        those img arrays every time somebody requests the image data by get_img_data() and img_data.
        """

        # check that all data has the same dimensions
        if self._background_data is not None:
            if self._img_data.shape != self._background_data.shape:
                self._background_data = None
        if self._img_corrections.has_items():
            if self._img_data.shape != self._img_corrections.shape:
                self._img_corrections.clear()
                self.transfer_correction.reset()
                self.corrections_removed.emit()

        # calculate the current _img_data
        if self._background_data is not None and not self._img_corrections.has_items():
            self._img_data_background_subtracted = self._img_data - (self._background_scaling *
                                                                     self._background_data +
                                                                     self._background_offset)
        elif self._background_data is None and self._img_corrections.has_items():
            self._img_data_absorption_corrected = self._img_data / self._img_corrections.get_data()

        elif self._background_data is not None and self._img_corrections.has_items():
            self._img_data_background_subtracted_absorption_corrected = (self._img_data - (
                    self._background_scaling * self._background_data + self._background_offset)) / \
                                                                        self._img_corrections.get_data()

    @property
    def img_data(self):
        """
        :return:
            The image based on the current state of the ImgData object. It will apply all image correction as well as
            background subtraction. in case you want the raw data without corrections, please use the
            raw_img_data property.
        """
        # if self._img_data is None:
        #     return None

        if self._background_data is None and not self._img_corrections.has_items():
            return self._img_data * self.factor

        elif self._background_data is not None and not self._img_corrections.has_items():
            return self._img_data_background_subtracted * self.factor

        elif self._background_data is None and self._img_corrections.has_items():
            return self._img_data_absorption_corrected * self.factor

        elif self._background_data is not None and self._img_corrections.has_items():
            return self._img_data_background_subtracted_absorption_corrected * self.factor

    @property
    def raw_img_data(self):
        return self._img_data

    @property
    def untransformed_raw_img_data(self):
        self._reset_img_transformations()
        img_data = np.copy(self.raw_img_data)
        self._perform_img_transformations()
        return img_data

    def rotate_img_p90(self):
        """
        Rotates the image by 90 degree and updates the background accordingly (does not effect absorption correction).
        The transformation is saved and applied to every new image and background image loaded.
        The img_changed signal will be emitted after the process.
        """
        self._img_data = rotate_matrix_p90(self._img_data)

        if self._background_data is not None:
            self._background_data = rotate_matrix_p90(self._background_data)

        self.img_transformations.append(rotate_matrix_p90)

        self.transformations_changed.emit()
        self._calculate_img_data()
        self.img_changed.emit()

    def rotate_img_m90(self):
        """
        Rotates the image by -90 degree and updates the background accordingly (does not effect absorption correction).
        The transformation is saved and applied to every new image and background image loaded.
        The img_changed signal will be emitted after the process.
        """
        self._img_data = rotate_matrix_m90(self._img_data)
        if self._background_data is not None:
            self._background_data = rotate_matrix_m90(self._background_data)
        self.img_transformations.append(rotate_matrix_m90)
        self.transformations_changed.emit()

        self._calculate_img_data()
        self.img_changed.emit()

    def flip_img_horizontally(self):
        """
        Flips image about a horizontal axis and updates the background accordingly (does not effect absorption
        correction). The transformation is saved and applied to every new image and background image loaded.
        The img_changed signal will be emitted after the process.
        """
        self._img_data = np.fliplr(self._img_data)
        if self._background_data is not None:
            self._background_data = np.fliplr(self._background_data)
        self.img_transformations.append(np.fliplr)
        self.transformations_changed.emit()

        self._calculate_img_data()
        self.img_changed.emit()

    def flip_img_vertically(self):
        """
        Flips image about a vertical axis and updates the background accordingly (does not effect absorption
        correction). The transformation is saved and applied to every new image and background image loaded.
        The img_changed signal will be emitted after the process.
        """
        self._img_data = np.flipud(self._img_data)
        if self._background_data is not None:
            self._background_data = np.flipud(self._background_data)
        self.img_transformations.append(np.flipud)
        self.transformations_changed.emit()

        self._calculate_img_data()
        self.img_changed.emit()

    def reset_transformations(self, img_changed=True):
        """
        Reverts all image transformations and resets the transformation stack.
        The img_changed signal will be emitted after the process, if set to true.
        """
        self._reset_img_transformations()
        self._reset_background_transformations()

        self.img_transformations = []
        self.transformations_changed.emit()

        self._calculate_img_data()
        if img_changed:
            self.img_changed.emit()

    def _reset_img_transformations(self):
        for transformation in reversed(self.img_transformations):
            if transformation == rotate_matrix_p90:
                self._img_data = rotate_matrix_m90(self._img_data)
            elif transformation == rotate_matrix_m90:
                self._img_data = rotate_matrix_p90(self._img_data)
            else:
                self._img_data = transformation(self._img_data)

    def _reset_background_transformations(self):
        if self._background_data is None:
            return

        for transformation in reversed(self.img_transformations):
            if transformation == rotate_matrix_p90:
                self._background_data = rotate_matrix_m90(self._background_data)
            elif transformation == rotate_matrix_m90:
                self._background_data = rotate_matrix_p90(self._background_data)
            else:
                self._background_data = transformation(self._background_data)

    def _perform_img_transformations(self):
        """
        Performs all saved image transformation on original image.
        """
        for transformation in self.img_transformations:
            self._img_data = transformation(self._img_data)

    def _perform_background_transformations(self):
        """
        Performs all saved image transformation on background image.
        """
        if self._background_data is not None:
            for transformation in self.img_transformations:
                self._background_data = transformation(self._background_data)

    def get_transformations_string_list(self):
        transformation_list = []
        for transformation in self.img_transformations:
            transformation_list.append(transformation.__name__)
        return transformation_list

    def load_transformations_string_list(self, transformations):
        self._reset_img_transformations()
        self._reset_background_transformations()
        self.img_transformations = []
        for transformation in transformations:
            if transformation == "flipud":
                self.img_transformations.append(np.flipud)
            elif transformation == "fliplr":
                self.img_transformations.append(np.fliplr)
            elif transformation == "rotate_matrix_m90":
                self.img_transformations.append(rotate_matrix_m90)
            elif transformation == "rotate_matrix_p90":
                self.img_transformations.append(rotate_matrix_p90)
        self._perform_img_transformations()
        self._perform_background_transformations()

    def add_img_correction(self, correction, name=None):
        """
        Adds a correction to be applied to the image. Corrections are applied multiplicative for each pixel and after
        each other, depending on the order of addition.
        :param external:
        :param correction: An Object inheriting the ImgCorrectionInterface.
        :type correction: ImgCorrectionInterface
        :param name: correction can be given a name, to selectively delete or obtain later.
        :type name: str
        """
        self._img_corrections.add(correction, name)
        self._calculate_img_data()
        self.img_changed.emit()

    def get_img_correction(self, name):
        """
        :param name: correction name which was specified during the addition of the image correction.
        :return: the specified correction
        """
        return self._img_corrections.get_correction(name)

    def delete_img_correction(self, name=None):
        """
        :param name: deletes a correction from the correction calculation with a specific name. if no name is specified
         the last added correction is deleted.
        """
        self._img_corrections.delete(name)
        self._calculate_img_data()
        self.img_changed.emit()

    def enable_transfer_function(self):
        if self.transfer_correction.get_data() is not None and \
                self.get_img_correction('transfer') is None:
            self.add_img_correction(self.transfer_correction, 'transfer')
        if self.get_img_correction('transfer') is not None:
            self._calculate_img_data()
            self.img_changed.emit()

    def disable_transfer_function(self):
        if self.get_img_correction('transfer') is not None:
            self.delete_img_correction('transfer')

    @property
    def img_corrections(self):
        return self._img_corrections

    def has_corrections(self):
        """
        :return: Whether the ImgData object has active absorption corrections or not
        """
        return self._img_corrections.has_items()

    def _get_file_info(self, image):
        """
        reads the file info from tif_tags and returns a file info
        """
        result = ""
        end_result = ""
        tags = image.tag
        useful_keys = []
        for key in tags.keys():
            if key > 300:
                useful_keys.append(key)

        useful_keys.sort()
        for key in useful_keys:
            tag = tags[key][0]
            if isinstance(tag, str):
                new_line = str(tag) + "\n"
                new_line = new_line.replace(":", ":\t", 1)
                if 'TIFFImageDescription' in new_line:
                    end_result = new_line
                else:
                    result += new_line
        return result + end_result

    def _get_motors_info(self, image):
        """
        reads the file info from tif_tags and returns positions of vertical, horizontal, focus and omega motors
        """
        result = {}
        tags = image.tag

        useful_tags = ['Horizontal:', 'Vertical:', 'Focus:', 'Omega:']

        try:
            tag_values = tags.itervalues()
        except AttributeError:
            tag_values = tags.values()

        for value in tag_values:
            for key in useful_tags:
                if key in str(value):
                    k, v = str(value[0]).split(':')
                    result[str(k)] = float(v)
        return result

    @property
    def autoprocess(self):
        return self._autoprocess

    @autoprocess.setter
    def autoprocess(self, new_val):
        self._autoprocess = new_val
        if new_val:
            self._directory_watcher.activate()
        else:
            self._directory_watcher.deactivate()

    @property
    def factor(self):
        return self._factor

    @factor.setter
    def factor(self, new_value):
        self._factor = new_value
        self.img_changed.emit()

    def blockSignals(self, block=True):
        for member in vars(self):
            attr = getattr(self, member)
            if isinstance(attr, Signal):
                attr.blocked = block


class BackgroundDimensionWrongException(Exception):
    pass
